import logging
import pickle
from dataclasses import dataclass
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from matplotlib.axes import Axes
from matplotlib.colors import Colormap, LogNorm, Normalize
from matplotlib.figure import Figure
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import MinMaxScaler

import flyvis
from flyvis.utils.activity_utils import CentralActivity

from .stimulus_responses import naturalistic_stimuli_responses
from .visualization import plt_utils
from .visualization.plt_utils import check_markers

__all__ = ["Embedding", "Clustering", "GaussianMixtureClustering", "EmbeddingPlot"]

INVALID_INT = -99999

logging = logging.getLogger(__name__)
if TYPE_CHECKING:
    from umap import UMAP
else:
    UMAP = TypeVar("UMAP")


@dataclass
class Embedding:
    """
    Embedding of the ensemble responses.

    Attributes:
        embedding (npt.NDArray): The embedded data.
        mask (npt.NDArray): Mask for valid data points.
        reducer (object): The reduction object used for embedding.
    """

    embedding: npt.NDArray = None
    mask: npt.NDArray = None
    reducer: object = None

    @property
    def cluster(self) -> "Clustering":
        """Returns a Clustering object for this embedding."""
        return Clustering(self)

    @property
    def embedding(self) -> npt.NDArray:  # noqa: F811
        """Returns the embedded data."""
        return getattr(self, "_embedding", None)

    @embedding.setter
    def embedding(self, value: npt.NDArray) -> None:
        """
        Sets the embedding and scales it to range (0, 1).

        Args:
            value: The embedding array to set.
        """
        self._embedding, self.minmaxscaler = scale_tensor(value)

    def plot(
        self,
        fig: Figure = None,
        ax: Axes = None,
        figsize: tuple = None,
        plot_mode: str = "paper",
        fontsize: int = 5,
        colors: npt.NDArray = None,
        **kwargs,
    ) -> tuple[Figure, Axes]:
        """
        Plot the embedding.

        Args:
            fig: Existing figure to plot on.
            ax: Existing axes to plot on.
            figsize: Size of the figure.
            plot_mode: Mode for plotting ('paper', 'small', or 'large').
            fontsize: Font size for annotations.
            colors: Colors for data points.
            **kwargs: Additional arguments passed to plot_embedding.

        Returns:
            A tuple containing the figure and axes objects.

        Raises:
            AssertionError: If the embedding is not 2-dimensional.
        """
        if self.embedding.shape[1] != 2:
            raise AssertionError("Embedding must be 2-dimensional for plotting")
        if figsize is None:
            figsize = [0.94, 2.38]
        return plot_embedding(
            self.embedding,
            colors=colors,
            task_error=None,
            labels=None,
            mask=self.mask,
            fit_gaussians=False,
            annotate=False,
            title="",
            fig=fig,
            ax=ax,
            mode=plot_mode,
            figsize=figsize,
            fontsize=fontsize,
            **kwargs,
        )


def scale_tensor(tensor: npt.NDArray) -> tuple[npt.NDArray, MinMaxScaler]:
    """
    Scale tensor to range (0, 1).

    Args:
        tensor: Input tensor to be scaled.

    Returns:
        A tuple containing the scaled tensor and the MinMaxScaler object.
    """
    s = MinMaxScaler(feature_range=(0, 1), copy=True, clip=False)
    return s.fit_transform(tensor), s


@dataclass
class GaussianMixtureClustering:
    """
    Gaussian Mixture Clustering of the embeddings.

    Attributes:
        embedding (Embedding): The embedding to cluster.
        range_n_clusters (Iterable[int]): Range of number of clusters to try.
        n_init (int): Number of initializations for GMM.
        max_iter (int): Maximum number of iterations for GMM.
        random_state (int): Random state for reproducibility.
        labels (npt.NDArray): Cluster labels.
        gm (object): Fitted GaussianMixture object.
        scores (list): Scores for each number of clusters.
        n_clusters (list): Number of clusters tried.
    """

    embedding: Embedding = None
    range_n_clusters: Iterable[int] = None
    n_init: int = 1
    max_iter: int = 1000
    random_state: int = 0
    labels: npt.NDArray = None
    gm: object = None
    scores: list = None
    n_clusters: list = None

    def __call__(
        self,
        range_n_clusters: Iterable[int] = None,
        n_init: int = 1,
        max_iter: int = 1000,
        random_state: int = 0,
        **kwargs,
    ) -> "GaussianMixtureClustering":
        """
        Perform Gaussian Mixture clustering.

        Args:
            range_n_clusters: Range of number of clusters to try.
            n_init: Number of initializations for GMM.
            max_iter: Maximum number of iterations for GMM.
            random_state: Random state for reproducibility.
            **kwargs: Additional arguments for gaussian_mixture function.

        Returns:
            Self with updated clustering results.
        """
        self.labels, self.gm, self.scores, self.n_clusters = gaussian_mixture(
            self.embedding.embedding,
            self.embedding.mask,
            range_n_clusters=range_n_clusters,
            n_init=n_init,
            max_iter=max_iter,
            random_state=random_state,
            **kwargs,
        )
        self.range_n_clusters = range_n_clusters
        self.n_init = n_init
        self.max_iter = max_iter
        self.random_state = random_state
        self.kwargs = kwargs
        return self

    def task_error_sort_labels(self, task_error: npt.NDArray, mode: str = "mean") -> None:
        """
        Sort cluster labels based on task error.

        Args:
            task_error: Array of task errors.
            mode: Method to compute task error ('mean', 'min', or 'median').
        """
        self.labels = task_error_sort_labels(task_error, self.labels, mode=mode)

    def plot(
        self,
        task_error: npt.NDArray = None,
        colors: npt.NDArray = None,
        annotate: bool = True,
        annotate_scores: bool = False,
        fig: Figure = None,
        ax: Axes = None,
        figsize: tuple = None,
        plot_mode: str = "paper",
        fontsize: int = 5,
        **kwargs,
    ) -> "EmbeddingPlot":
        """
        Plot the clustering results.

        Args:
            task_error: Array of task errors.
            colors: Colors for data points.
            annotate: Whether to annotate clusters.
            annotate_scores: Whether to annotate BIC scores.
            fig: Existing figure to plot on.
            ax: Existing axes to plot on.
            figsize: Size of the figure.
            plot_mode: Mode for plotting ('paper', 'small', or 'large').
            fontsize: Font size for annotations.
            **kwargs: Additional arguments for plot_embedding function.

        Returns:
            An EmbeddingPlot object.

        Raises:
            AssertionError: If the embedding is not 2-dimensional.
        """
        if self.embedding.embedding.shape[1] != 2:
            raise AssertionError("Embedding must be 2-dimensional for plotting")
        if figsize is None:
            figsize = [0.94, 2.38]
        fig, ax = plot_embedding(
            self.embedding.embedding,
            colors=colors,
            task_error=task_error,
            labels=self.labels,
            gm=self.gm,
            mask=self.embedding.mask,
            fit_gaussians=True,
            annotate=annotate,
            title="",
            fig=fig,
            ax=ax,
            mode=plot_mode,
            figsize=figsize,
            fontsize=fontsize,
            range_n_clusters=self.range_n_clusters,
            n_init_gaussian_mixture=self.n_init,
            gm_kwargs=self.kwargs,
            **kwargs,
        )
        if annotate_scores:
            ax.annotate(
                "BIC: {:.2f}".format(np.min(self.scores)),
                xy=(ax.get_xlim()[0], ax.get_ylim()[1]),
                ha="left",
                va="top",
                fontsize=fontsize,
            )
        return EmbeddingPlot(fig, ax, None, None, self.gm.n_components, self)


@dataclass
class EmbeddingPlot:
    fig: Figure = None
    ax: Axes = None
    cmap: Colormap = None
    norm: Normalize = None
    n_clusters: int = None
    cluster: GaussianMixtureClustering = None


@dataclass
class Clustering:
    """Clustering of the embedding.

    Attributes:
        embedding (Embedding): The embedding to be clustered.
    """

    embedding: Embedding = None

    @property
    def gaussian_mixture(self) -> GaussianMixtureClustering:
        """Create a GaussianMixtureClustering object for the embedding.

        Returns:
            GaussianMixtureClustering: A clustering object for Gaussian mixture models.
        """
        return GaussianMixtureClustering(self.embedding)


def gaussian_mixture(
    X: np.ndarray,
    mask: np.ndarray,
    range_n_clusters: Optional[Union[List[int], np.ndarray]] = None,
    n_init: int = 1,
    max_iter: int = 1000,
    random_state: int = 0,
    criterion: str = "bic",
    **kwargs,
) -> Tuple[np.ndarray, GaussianMixture, np.ndarray, np.ndarray]:
    """Fit Gaussian Mixtures to the data.

    Args:
        X: Input data with shape (n_samples, n_features).
        mask: Boolean mask for valid samples.
        range_n_clusters: Range of number of components to fit.
        n_init: Number of initializations for each number of components.
        max_iter: Maximum number of iterations for the fitting process.
        random_state: Random state for reproducibility.
        criterion: Criterion to use for selecting the number of components.
            Options are "bic", "aic", or "score".
        **kwargs: Additional keyword arguments for GaussianMixture.

    Returns:
        A tuple containing:

        - labels: Cluster labels for each sample.
        - gm: Fitted GaussianMixture object.
        - metric: Metric values (BIC, AIC, or score) for each number of components.
        - range_n_clusters: Range of number of components tried.

    Raises:
        ValueError: If an unknown criterion is provided.
    """
    if range_n_clusters is None:
        range_n_clusters = np.array([1, 2, 3, 4, 5])

    if isinstance(range_n_clusters, list):
        range_n_clusters = np.array(range_n_clusters)

    n_samples = X.shape[0]
    labels = np.ones([n_samples], dtype=int) * INVALID_INT
    X = X[mask]

    n_valid_samples = X.shape[0]
    range_n_clusters = range_n_clusters[range_n_clusters <= n_valid_samples]

    scores = []
    aic = []
    bic = []
    for n_clusters in range_n_clusters:
        gm = GaussianMixture(
            n_components=n_clusters,
            random_state=random_state,
            max_iter=max_iter,
            n_init=n_init,
            **kwargs,
        )
        gm.fit(X)
        scores.append(-gm.score(X))
        aic.append(gm.aic(X))
        bic.append(gm.bic(X))

    if criterion == "bic":
        metric = bic
    elif criterion == "aic":
        metric = aic
    elif criterion == "score":
        metric = scores
    else:
        raise ValueError(f"Unknown criterion: {criterion}")

    n_components = range_n_clusters[np.argmin(metric)]
    gm = GaussianMixture(
        n_components=n_components,
        random_state=random_state,
        max_iter=max_iter,
        n_init=n_init,
        **kwargs,
    )

    labels[mask] = gm.fit_predict(X).astype(int)

    return labels, gm, metric, range_n_clusters


class EnsembleEmbedding:
    """Embedding of the ensemble responses.

    Args: responses (CentralActivity): CentralActivity object
    """

    def __init__(self, responses: CentralActivity):
        self.responses = responses

    def from_cell_type(
        self,
        cell_type,
        embedding_kwargs=None,
    ) -> Embedding:
        """Umap Embedding of the responses of a specific cell type."""

        embedding_kwargs = embedding_kwargs or {}
        return Embedding(*umap_embedding(self.responses[cell_type], **embedding_kwargs))

    def __call__(
        self,
        arg: Union[str, Iterable],
        embedding_kwargs=None,
    ):
        if isinstance(arg, str):
            return self.from_cell_type(arg, embedding_kwargs)
        else:
            raise ValueError("arg")


def umap_embedding(
    X: np.ndarray,
    n_neighbors: int = 5,
    min_dist: float = 0.12,
    spread: float = 9.0,
    random_state: int = 42,
    n_components: int = 2,
    metric: str = "correlation",
    n_epochs: int = 1500,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray, Optional[UMAP]]:
    """
    Perform UMAP embedding on input data.

    Args:
        X: Input data with shape (n_samples, n_features).
        n_neighbors: Number of neighbors to consider for each point.
        min_dist: Minimum distance between points in the embedding space.
        spread: Determines how spread out all embedded points are overall.
        random_state: Random seed for reproducibility.
        n_components: Number of dimensions in the embedding space.
        metric: Distance metric to use.
        n_epochs: Number of training epochs for embedding optimization.
        **kwargs: Additional keyword arguments for UMAP.

    Returns:
        A tuple containing:
        - embedding: The UMAP embedding (n_samples, n_components). May be NaN
          if insufficient data.
        - mask: Boolean mask (length n_samples). When reducer is not None,
          True indicates rows with nonzero variance that were also connected
          in the UMAP graph. When reducer is None (insufficient data), True
          indicates only rows with nonzero variance.
        - reducer: The fitted UMAP object or None if fewer than 2 rows had
          nonzero variance.

    Raises:
        ValueError: If n_components is too large relative to sample size.

    Note:
        This function handles reshaping of input data and removes constant rows.
    """
    # umap import would slow down whole library import
    from umap import UMAP
    from umap.utils import disconnected_vertices

    if n_components > X.shape[0] - 2:
        raise ValueError(
            "number of components must be 2 smaller than sample size. "
            "See: https://github.com/lmcinnes/umap/issues/201"
        )

    if len(X.shape) > 2:
        shape = X.shape
        X = X.reshape(X.shape[0], -1)
        logging.info("reshaped X from %s to %s", shape, X.shape)

    n_samples = X.shape[0]
    embedding = np.ones([n_samples, n_components]) * np.nan
    # umap doesn't like constant rows
    mask = ~np.isclose(X.std(axis=1), 0)
    X_nonconst = X[mask]

    # If fewer than 2 rows remain, skip UMAP and return embedding of NaNs.
    if X_nonconst.shape[0] < 2:
        return embedding, mask, None

    reducer = UMAP(
        n_neighbors=n_neighbors,
        min_dist=min_dist,
        random_state=random_state,
        n_components=n_components,
        metric=metric,
        spread=spread,
        n_epochs=n_epochs,
        **kwargs,
    )
    _embedding = reducer.fit_transform(X_nonconst)

    # gaussian mixture doesn't like nans through disconnected vertices in umap
    connected_vertices_mask = ~disconnected_vertices(reducer)
    mask[mask] = mask[mask] & connected_vertices_mask
    embedding[mask] = _embedding[connected_vertices_mask]
    return embedding, mask, reducer


def plot_embedding(
    X: np.ndarray,
    colors: Optional[np.ndarray] = None,
    task_error: Optional[np.ndarray] = None,
    labels: Optional[np.ndarray] = None,
    gm: Optional[GaussianMixture] = None,
    mask: Optional[np.ndarray] = None,
    fit_gaussians: bool = True,
    annotate: bool = True,
    contour_gaussians: bool = True,
    range_n_clusters: List[int] = [1, 2, 3, 4, 5],
    n_init_gaussian_mixture: int = 10,
    title: str = "",
    fig: Optional[Figure] = None,
    ax: Optional[Axes] = None,
    mode: Optional[str] = None,
    figsize: List[float] = [3, 3],
    s: float = 20,
    fontsize: int = 5,
    ax_lim_pad: float = 0.2,
    task_error_sort_mode: str = "mean",
    err_x_offset: float = 0.025,
    err_y_offset: float = -0.025,
    gm_kwargs: Optional[Dict] = None,
) -> Tuple[Figure, Axes]:
    """
    Plot the embedding of data points with optional Gaussian mixture clustering.

    Args:
        X: Input data with shape (n_samples, 2).
        colors: Color values for each data point.
        task_error: Task error values for each data point.
        labels: Cluster labels for each data point.
        gm: Fitted GaussianMixture object.
        mask: Boolean mask for valid samples.
        fit_gaussians: Whether to fit Gaussian mixtures.
        annotate: Whether to annotate clusters with performance metrics.
        contour_gaussians: Whether to plot contours for Gaussian mixtures.
        range_n_clusters: Range of number of clusters to try.
        n_init_gaussian_mixture: Number of initializations for Gaussian mixture.
        title: Title of the plot.
        fig: Existing figure to plot on.
        ax: Existing axes to plot on.
        mode: Plotting mode ('paper', 'small', or 'large').
        figsize: Size of the figure.
        s: Size of scatter points.
        fontsize: Font size for annotations.
        ax_lim_pad: Padding for axis limits.
        task_error_sort_mode: Mode for sorting task errors ('mean' or 'min').
        err_x_offset: X-offset for error annotations.
        err_y_offset: Y-offset for error annotations.
        gm_kwargs: Additional keyword arguments for Gaussian mixture.

    Returns:
        A tuple containing the figure and axes objects.

    Note:
        This function handles various plotting scenarios including Gaussian mixture
        clustering and task error annotation.
    """
    if mask is None:
        mask = slice(None)

    if colors is None:
        colors = np.array(X.shape[0] * ["#779eaa"])
    colors = colors[mask]

    if task_error is not None:
        task_error = task_error[mask]

    if mode == "paper":
        figsize = [2.68, 2.38]
        s = 20
        fontsize = 5
    if mode == "small":
        figsize = [3, 3]
        s = 20
        fontsize = 5
    elif mode == "large":
        figsize = [10, 10]
        s = 60
        fontsize = 12

    fig, ax = plt_utils.init_plot(figsize=figsize, fontsize=fontsize, fig=fig, ax=ax)

    xc = X[:, 0]
    yc = X[:, 1]
    x_min, x_max = plt_utils.get_lims(xc, ax_lim_pad)
    y_min, y_max = plt_utils.get_lims(yc, ax_lim_pad)

    if fit_gaussians:
        # breakpoint()
        if labels is not None and gm:
            _labels = labels
        elif labels is not None:
            _labels, gm, scores = gaussian_mixture(
                X,
                mask,
                range_n_clusters,
                n_init=n_init_gaussian_mixture,
                **(gm_kwargs or {}),
            )
        if labels is not None:
            assert (_labels == labels).all()
        X = X[mask]
        labels = _labels[mask]

        if task_error is not None:
            labels = task_error_sort_labels(task_error, labels, mode=task_error_sort_mode)
        else:
            # because the sorting is later based on this to plot better
            # performing models on top of worse performing models
            task_error = np.arange(X.shape[0])
            annotate = False

        if contour_gaussians:
            x = np.linspace(x_min, x_max, 1000)
            y = np.linspace(y_min, y_max, 1000)
            _X, _Y = np.meshgrid(x, y)
            XX = np.array([_X.ravel(), _Y.ravel()]).T
            Z = np.exp(gm.score_samples(XX))
            Z /= Z.sum()
            Z = Z.reshape(_X.shape)
            # print(Z.min(), Z.max())
            ax.contourf(
                _X,
                _Y,
                Z,
                norm=LogNorm(vmin=1e-25, vmax=1e0),
                levels=np.logspace(-25, 0, 15),
                cmap=plt.cm.binary,
                alpha=0.1,
                zorder=-1,
            )

        MARKERS = check_markers(len(np.unique(labels)))
        for label in np.unique(labels).astype(int):
            # to plot best performing models on top
            _argsort = np.argsort(task_error)[::-1]
            _X = X[_argsort]
            _colors = colors[_argsort]
            _labels = labels[_argsort]

            ax.scatter(
                _X[_labels == label, 0],
                _X[_labels == label, 1],
                marker=MARKERS[label],
                c=_colors[_labels == label],
                s=s,
                alpha=0.6,
                edgecolors="none",
            )

            if annotate:
                # annotate performance average
                x, y = X[labels == label, 0].mean(), X[labels == label, 1].mean()
                if task_error_sort_mode == "mean":
                    _perf = task_error[labels == label].mean().item()
                elif task_error_sort_mode == "min":
                    _perf = task_error[labels == label].min().item()
                else:
                    raise ValueError()

                ax.annotate(
                    f"{_perf:.3f}",
                    (
                        x + err_x_offset * (x_max - x_min),
                        y + err_y_offset * (y_max - y_min),
                    ),
                    fontsize=fontsize,
                    ha="left",
                    va="top",
                )

    elif labels is not None:
        X = X[mask]
        labels = labels[mask]
        MARKERS = check_markers(len(np.unique(labels)))
        for label in np.unique(labels).astype(int):
            ax.scatter(
                X[labels == label, 0],
                X[labels == label, 1],
                marker=MARKERS[label],
                c=colors[labels == label],
                s=s,
                alpha=0.6,
                edgecolors="none",
            )

            if annotate and task_error is not None:
                # annotate performance average
                x, y = X[labels == label, 0].mean(), X[labels == label, 1].mean()
                _perf = task_error[labels == label].mean().item()
                ax.annotate(
                    f"{_perf:.3G}",
                    (
                        x + err_x_offset * (x_max - x_min),
                        y - err_y_offset * (y_max - y_min),
                    ),
                    fontsize=fontsize,
                    ha="left",
                    va="top",
                )
    else:
        X = X[mask]
        ax.scatter(X[:, 0], X[:, 1], c=colors, s=s, alpha=0.6, edgecolors="none")

    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)

    ax.set_title(title, fontsize=fontsize)
    plt_utils.rm_spines(ax)

    return fig, ax


def task_error_sort_labels(
    task_error: np.ndarray, labels: np.ndarray, mode: str = "mean"
) -> np.ndarray:
    """
    Permute cluster label IDs based on task error.

    Args:
        task_error: Array of task errors for each sample.
        labels: Array of cluster labels.
        mode: Method to compute task error. Options: "mean", "min", "median".

    Returns:
        Array of sorted labels.

    Raises:
        ValueError: If an invalid mode is provided.

    Note:
        Sorts labels so the best performing cluster has label 0, second best 1, etc.
    """
    sorted_labels = np.ones_like(labels, dtype=int) * INVALID_INT

    task_error_cluster = []
    for label_id in np.unique(labels[labels != INVALID_INT]):
        if mode == "mean":
            _error = task_error[labels == label_id].mean()
        elif mode == "min":
            _error = task_error[labels == label_id].min()
        elif mode == "median":
            _error = np.median(task_error[labels == label_id])
        else:
            raise ValueError("Invalid mode for task error calculation")
        task_error_cluster.append(_error)

    for i, x in enumerate(np.argsort(task_error_cluster)):
        sorted_labels[np.isclose(labels, x)] = i

    return sorted_labels


def get_cluster_to_indices(
    mask: np.ndarray, labels: np.ndarray, task_error: Optional[np.ndarray] = None
) -> Dict[int, np.ndarray]:
    """
    Map cluster labels to corresponding indices.

    Args:
        mask: Boolean mask for valid samples.
        labels: Array of cluster labels.
        task_error: Optional array of task errors for sorting labels.

    Returns:
        Dictionary mapping cluster labels to arrays of indices.
    """
    indices = np.arange(len(labels)).astype(int)
    if task_error is not None:
        labels = task_error_sort_labels(task_error.values, labels)

    indices = np.arange(len(labels))[mask]
    labels = labels[mask]
    cluster_indices = {
        label_id: indices[labels == label_id] for label_id in np.unique(labels)
    }
    return cluster_indices


def compute_umap_and_clustering(
    ensemble: "flyvis.network.EnsembleView",
    cell_type: str,
    embedding_kwargs: Optional[Dict] = None,
    gm_kwargs: Optional[Dict] = None,
    subdir: str = "umap_and_clustering",
) -> GaussianMixtureClustering:
    """
    Compute UMAP embedding and Gaussian Mixture clustering of responses.

    Args:
        ensemble: EnsembleView object.
        cell_type: Type of cell to analyze.
        embedding_kwargs: UMAP embedding parameters.
        gm_kwargs: Gaussian Mixture clustering parameters.
        subdir: Subdirectory for storing results.

    Returns:
        GaussianMixtureClustering object.

    Note:
        Results are cached to disk for faster subsequent access.
    """
    if embedding_kwargs is None:
        embedding_kwargs = {
            "min_dist": 0.105,
            "spread": 9.0,
            "n_neighbors": 5,
            "random_state": 42,
            "n_epochs": 1500,
        }
    if gm_kwargs is None:
        gm_kwargs = {
            "range_n_clusters": [2, 3, 3, 4, 5],
            "n_init": 100,
            "max_iter": 1000,
            "random_state": 42,
            "tol": 0.001,
        }

    destination = ensemble.path / subdir

    def load_from_disk():
        with open((destination / cell_type).with_suffix(".pickle"), "rb") as f:
            embedding_and_clustering = pickle.load(f)

        logging.info("Loaded %s embedding and clustering from %s", cell_type, destination)
        return embedding_and_clustering

    if (destination / cell_type).with_suffix(".pickle").exists():
        return load_from_disk()

    def create_embedding_object(responses):
        central_responses = CentralActivity(
            responses['responses'].values, ensemble[0].connectome, keepref=True
        )
        embeddings = EnsembleEmbedding(central_responses)
        return embeddings

    responses = naturalistic_stimuli_responses(ensemble)
    embeddings = create_embedding_object(responses)

    embedding = embeddings.from_cell_type(cell_type, embedding_kwargs=embedding_kwargs)
    embedding_and_clustering = embedding.cluster.gaussian_mixture(**gm_kwargs)
    return embedding_and_clustering


@wraps(compute_umap_and_clustering)
def umap_and_clustering_generator(
    ensemble: "flyvis.network.EnsembleView", **kwargs
) -> Generator[Tuple[str, GaussianMixtureClustering], None, None]:
    """
    Generate UMAP and clustering for all cell types.

    Args:
        ensemble: EnsembleView object.
        **kwargs: Additional arguments for compute_umap_and_clustering.

    Yields:
        Tuple of cell type and corresponding GaussianMixtureClustering object.
    """
    for cell_type in ensemble[0].connectome_view.cell_types_sorted:
        yield cell_type, compute_umap_and_clustering(ensemble, cell_type, **kwargs)



import argparse
import shutil
from importlib.resources import files
from pathlib import Path


def get_valid_config_groups(pkg_config_dir):
    """Get list of valid config groups from package config directory."""
    config_path = Path(pkg_config_dir)
    valid_groups = []

    for path in config_path.rglob('*.yaml'):
        relative_path = path.relative_to(config_path)
        group = str(relative_path.parent / relative_path.stem)
        valid_groups.append(group)

    return sorted(valid_groups)


def init_config(args):
    """Initialize a local config directory with defaults from the package."""
    pkg_config_dir = str(files('flyvis') / 'config')
    output_path = Path(args.output_dir).resolve()
    success = False

    valid_groups = get_valid_config_groups(pkg_config_dir)

    if not args.config_groups:
        # Copy entire config structure contents to flyvis_config
        config_path = output_path / 'flyvis_config'
        config_path.mkdir(parents=True, exist_ok=True)

        # Copy contents of pkg_config_dir to flyvis_config
        for item in Path(pkg_config_dir).iterdir():
            dst = config_path / item.name
            if item.is_file():
                shutil.copy2(item, dst)
            else:
                shutil.copytree(item, dst, dirs_exist_ok=True)

        print(f"Copied full config structure to {config_path.absolute()}")
        print(
            "Remember: When creating custom configs, use different names than the "
            "defaults."
        )
        success = True
    else:
        # Copy only specified config groups
        for group in args.config_groups:
            # Strip leading 'config/' if present
            group = group[7:] if group.startswith('config/') else group
            src_path = Path(pkg_config_dir) / group
            if not src_path.exists():
                print(f"Warning: Config group '{group}' not found in package")
                print("\nValid config groups:")
                for valid_group in valid_groups:
                    print(f"  - {valid_group}")
                continue

            dst_path = output_path / group
            dst_path.parent.mkdir(parents=True, exist_ok=True)
            if src_path.is_file():
                shutil.copy2(src_path, dst_path)
            else:
                shutil.copytree(src_path, dst_path, dirs_exist_ok=True)

            print(f"Copied {group} to {dst_path.absolute()}.")
            print(
                "Remember: When creating custom configs, use different names than the "
                "defaults"
            )
            success = True

    if not success:
        print("\nNo valid config groups were copied. Available config groups:")
        for group in valid_groups:
            print(f"  - {group}")
        raise ValueError("No valid config groups were found to copy")

    # Update README content with correct path and location
    config_dir = '$(pwd)/flyvis_config' if not args.config_groups else '$(pwd)'
    readme_path = (
        output_path / ('flyvis_config' if not args.config_groups else '') / 'README.md'
    )
    readme_content = f"""# Custom FlyVis Configuration

This directory contains custom configurations for FlyVis. To use these configs:

Important: When creating custom configs, use different names than the defaults to
ensure proper override behavior.
For example:
- Default: syn_strength.yaml
- Custom: custom_syn_strength.yaml

Override specific configs:
```bash
flyvis train-single network/edge_config/syn_strength=custom_syn_strength --config-dir \
{config_dir}
```

The directory structure mirrors the package config structure, allowing you to override
specific parts while keeping others at their defaults.
Note that Hydra will look in the package's default config directory first, so your
custom configs must have different names.
"""

    readme_path.parent.mkdir(parents=True, exist_ok=True)
    with open(readme_path, 'w') as f:
        f.write(readme_content)

    print(
        f"\nConfig initialization complete! See {readme_path.absolute()} for usage "
        "instructions."
    )


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Initialize a local config directory with defaults from the package'
    )
    parser.add_argument(
        '--output-dir', default='.', help='Directory where config should be initialized'
    )
    parser.add_argument(
        '--config-groups',
        nargs='*',
        help=(
            'Specific config groups to initialize '
            '(e.g., network/edge_config/syn_strength). If none specified, copies all.'
        ),
    )

    args = parser.parse_args()
    init_config(args)



"""Animations of the Sintel data."""

from typing import Optional, Tuple

import numpy as np
from matplotlib import colormaps

from flyvis import utils

from ..visualization import figsize_utils
from .animations import AnimationCollector
from .hexflow import HexFlow
from .hexscatter import HexScatter


class SintelSample(AnimationCollector):
    """Sintel-specific animation of input, target, and groundtruth data.

    Args:
        lum: Input of shape (n_samples, n_frames, n_hexals).
        target: Target of shape (n_samples, n_frames, n_dims, n_features).
        prediction: Optional prediction of shape
            (n_samples, n_frames, n_dims, n_features).
        target_cmap: Colormap for the target (depth).
        fontsize: Font size for labels and titles.
        labelxy: Normalized x and y location of the label.
        max_figure_height_cm: Maximum figure height in centimeters.
        panel_height_cm: Height of each panel in centimeters.
        max_figure_width_cm: Maximum figure width in centimeters.
        panel_width_cm: Width of each panel in centimeters.
        title1: Title for the input panel.
        title2: Title for the target panel.
        title3: Title for the prediction panel.

    Attributes:
        fig (Figure): Matplotlib figure instance.
        axes (List[Axes]): List of matplotlib axes instances.
        lum (np.ndarray): Input data.
        target (np.ndarray): Target data.
        prediction (Optional[np.ndarray]): Prediction data.
        extent (Tuple[float, float, float, float]): Extent of the hexagonal grid.
        n_samples (int): Number of samples.
        frames (int): Number of frames.
        update (bool): Whether to update the canvas after an animation step.
        labelxy (Tuple[float, float]): Normalized x and y location of the label.
        animations (List): List of animation objects.
        batch_sample (int): Batch sample to start from.
    """

    def __init__(
        self,
        lum: np.ndarray,
        target: np.ndarray,
        prediction: Optional[np.ndarray] = None,
        target_cmap: str = colormaps["binary_r"],
        fontsize: float = 5,
        labelxy: Tuple[float, float] = (-0.1, 1),
        max_figure_height_cm: float = 22,
        panel_height_cm: float = 3,
        max_figure_width_cm: float = 18,
        panel_width_cm: float = 3.6,
        title1: str = "input",
        title2: str = "target",
        title3: str = "prediction",
    ) -> None:
        figsize = figsize_utils.figsize_from_n_items(
            2 if prediction is None else 3,
            max_figure_height_cm=max_figure_height_cm,
            panel_height_cm=panel_height_cm,
            max_figure_width_cm=max_figure_width_cm,
            panel_width_cm=panel_width_cm,
        )
        self.fig, self.axes = figsize.axis_grid(
            hspace=0.0,
            wspace=0,
            fontsize=fontsize,
            unmask_n=2 if prediction is None else 3,
        )

        self.lum = lum
        self.target = target
        self.prediction = prediction
        self.extent = utils.hex_utils.get_hextent(self.lum.shape[-1])

        self.n_samples, self.frames = self.lum.shape[0:2]
        self.update = False
        self.labelxy = labelxy

        animations = []
        animations.append(
            HexScatter(
                self.lum,
                fig=self.fig,
                ax=self.axes[0],
                title=title1,
                edgecolor=None,
                update_edge_color=True,
                fontsize=fontsize,
                cbar=True,
                labelxy=labelxy,
            )
        )

        if self.target.shape[-2] == 2:
            animations.append(
                HexFlow(
                    flow=self.target,
                    fig=self.fig,
                    ax=self.axes[1],
                    cwheel=True,
                    cwheelxy=(-0.7, 0.7),
                    title=title2,
                    label="",
                    fontsize=fontsize,
                )
            )
            if prediction is not None:
                animations.append(
                    HexFlow(
                        flow=self.prediction,
                        fig=self.fig,
                        ax=self.axes[2],
                        cwheel=True,
                        cwheelxy=(-0.7, 0.7),
                        title=title3,
                        label="",
                        fontsize=fontsize,
                    )
                )
        else:
            animations.append(
                HexScatter(
                    self.target,
                    fig=self.fig,
                    ax=self.axes[1],
                    cmap=target_cmap,
                    title=title2,
                    edgecolor=None,
                    fontsize=fontsize,
                    cbar=True,
                    labelxy=labelxy,
                )
            )
            if prediction is not None:
                animations.append(
                    HexScatter(
                        self.prediction,
                        fig=self.fig,
                        ax=self.axes[2],
                        cmap=target_cmap,
                        title=title3,
                        edgecolor=None,
                        fontsize=fontsize,
                        cbar=True,
                        labelxy=labelxy,
                    )
                )
        self.animations = animations
        self.batch_sample = 0
        super().__init__(None, self.fig)



"""
The parameters that the networks can be initialized with. Each parameter is a
type on its own, because different parameters are shared differently. These
types handle the initialization of indices to perform gather and scatter opera-
tions. Parameter types can be initialized from a range of initial distribution
types.
"""

import functools
import logging
from copy import deepcopy
from typing import Any, Dict, Iterable, List

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from datamate import Namespace
from torch import Tensor

from flyvis import device
from flyvis.connectome import ConnectomeFromAvgFilters
from flyvis.utils.class_utils import forward_subclass
from flyvis.utils.tensor_utils import atleast_column_vector, where_equal_rows
from flyvis.utils.type_utils import byte_to_str

logging = logging.getLogger(__name__)

__all__ = [
    "Parameter",
    "RestingPotential",
    "TimeConstant",
    "SynapseSign",
    "SynapseCount",
    "SynapseCountScaling",
    "InitialDistribution",
    "Value",
    "Normal",
    "Lognormal",
]

# --- Initial distribution types ------------------------------------------------


class InitialDistribution:
    """Initial distribution base class.

    Attributes:
        raw_values (Tensor): Initial parameters must store raw_values as attribute in
            their __init__.
        readers (Dict[str, Tensor]): Readers will be written by the network during
            initialization.

    Note:
        To add a new initial distribution type, subclass this class and implement the
        __init__ method. The __init__ method should take the param_config as its first
        argument, and should store the attribute raw_values as a torch.nn.Parameter.

    Example:
        An example of a viable param_config is:
        ```python
        param_config = Namespace(
            requires_grad=True,
            initial_dist="Normal",
            mean=0,
            std=1,
            mode="sample",
        )
        ```
    """

    raw_values: Tensor
    readers: Dict[str, Tensor]

    @property
    def semantic_values(self):
        """Optional reparametrization of raw values invoked for computation."""
        return self.raw_values

    def __repr__(self):
        return f"{self.__class__.__name__} (semantic values): \n{self.semantic_values}"

    def __len__(self):
        return len(self.raw_values)

    def clamp(self, values, mode):
        """To clamp the raw_values of the parameters at initialization.

        Note, mild clash with raw_values/semantic_values reparametrization.
        Parameters that use reparametrization in terms of semantic_values
        should not use clamp.
        """
        if mode == "non_negative":
            values.clamp_(min=0)
        elif isinstance(mode, Iterable) and len(mode) == 2:
            values.clamp_(*mode)
        elif mode in [False, None]:
            return values
        else:
            raise ParameterConfigError(f"{mode} not a valid argument for clamp")
        return values


class Value(InitialDistribution):
    """Initializes parameters with a single value.

    Args:
        value: The value to initialize the parameter with.
        requires_grad (bool): Whether the parameter requires gradients.
        clamp (bool, optional): Whether to clamp the values. Defaults to False.
        **kwargs: Additional keyword arguments.

    Example:
        ```python
        param_config = Namespace(
            requires_grad=True,
            initial_dist="Value",
            value=0,
        )
        ```
    """

    def __init__(self, value, requires_grad, clamp=False, **kwargs) -> None:
        _values = torch.tensor(value).float()
        _values = self.clamp(_values, clamp)
        self.raw_values = nn.Parameter(_values, requires_grad=requires_grad)


class Normal(InitialDistribution):
    """Initializes parameters independently from normal distributions.

    Args:
        mean: The mean of the normal distribution.
        std: The standard deviation of the normal distribution.
        requires_grad (bool): Whether the parameter requires gradients.
        mode (str, optional): The initialization mode. Defaults to "sample".
        clamp (bool, optional): Whether to clamp the values. Defaults to False.
        seed (int, optional): Random seed for reproducibility. Defaults to None.
        **kwargs: Additional keyword arguments.

    Example:
        ```python
        param_config = Namespace(
            requires_grad=True,
            initial_dist="Normal",
            mean=0,
            std=1,
            mode="sample",
        )
        ```
    """

    def __init__(
        self, mean, std, requires_grad, mode="sample", clamp=False, seed=None, **kwargs
    ) -> None:
        if mode == "mean":
            _values = torch.tensor(mean).float()
        elif mode == "sample":
            # set seed for reproducibility and avoid seeding the global RNG
            generator = torch.Generator(device=device)
            if seed is not None:
                generator.manual_seed(seed)
            else:
                generator.seed()
            try:
                _values = torch.normal(
                    torch.tensor(mean).float(),
                    torch.tensor(std).float(),
                    generator=generator,
                )
            except RuntimeError as e:
                raise RuntimeError(
                    f"Failed to sample from normal with mean {mean} and std {std}"
                ) from e
        else:
            raise ValueError("Mode must be either mean or sample.")
        _values = self.clamp(_values, clamp)
        self.raw_values = nn.Parameter(_values, requires_grad=requires_grad)


class Lognormal(Normal):
    """Initializes parameters independently from lognormal distributions.

    Note:
        The lognormal distribution reparametrizes a normal through semantic values.

    Example:
        ```python
        param_config = Namespace(
            requires_grad=True,
            initial_dist="Lognormal",
            mean=0,
            std=1,
            mode="sample",
        )
        ```
    """

    @property
    def semantic_values(self):
        """n_syn ~ self._values.exp()."""
        return self.raw_values.exp()


# --- Parameter types -----------------------------------------------------------


def deepcopy_config(f):
    """Decorator to deepcopy the parameter configuration.

    Note:
        This decorator is necessary because the `__init__` method of parameter classes
        often modifies the `param_config` object. By creating a deep copy, we ensure
        that these modifications don't affect the original `param_config` object in the
        outer scope. This prevents unintended side effects and maintains the integrity
        of the original configuration.
    """

    @functools.wraps(f)
    def wrapper(cls, param_config, connectome):
        cls.config = deepcopy(param_config)
        return f(cls, deepcopy(param_config), connectome)

    return wrapper


class Parameter:
    """Base class for all parameters to share across nodes or edges.

    Args:
        param_config (Namespace): Namespace containing parameter configuration.
        connectome (Connectome): Connectome object.

    Attributes:
        parameter (InitialDistribution): InitialDistribution object.
        indices (torch.Tensor): Indices for parameter sharing.
        keys (List[Any]): Keys to access individual parameter values associated with
            certain identifiers.
        symmetry_masks (List[torch.Tensor]): Symmetry masks that can be configured
            optionally to apply further symmetry constraints to the parameter values.

    Note:
        Subclasses must implement `__init__(self, param_config, connectome_dir)` with the
        following requirements:

        1. Configure all attributes defined in the base class.
        2. Decorate `__init__` with `@deepcopy_config` if it updates `param_config` to
           prevent mutations in the outer scope.
        3. Update `param_config` with key-value pairs informed by `connectome` and
           matching the desired `InitialDistribution`.
        4. Store `parameter` from `InitialDistribution(param_config)`, which constructs
           and holds the `nn.Parameter`.
        5. Store `indices` for parameter sharing using
           `get_scatter_indices(dataframe, grouped_dataframe, groupby)`.
        6. Store `keys` to access individual parameter values associated with certain
           identifiers.
        7. Store `symmetry_masks` (optional) to apply further symmetry constraints to
           the parameter values.

        Example implementation structure:

        ```python
        @deepcopy_config
        def __init__(self, param_config: Namespace, connectome: Connectome):
            # Update param_config based on connectome data
            # ...

            # Initialize parameter
            self.parameter = InitialDistribution(param_config)

            # Set up indices, keys, and symmetry masks
            self.indices = get_scatter_indices(...)
            self.keys = ...
            self.symmetry_masks = symmetry_masks(...)
        ```
    """

    parameter: InitialDistribution
    indices: torch.Tensor
    symmetry_masks: List[torch.Tensor]
    keys: List[Any]

    @deepcopy_config
    def __init__(self, param_config: Namespace, connectome: ConnectomeFromAvgFilters):
        pass

    def __repr__(self):
        """Return a string representation of the Parameter object."""
        init_arg_names = list(self.__init__.__annotations__.keys())
        dir_type = self.__init__.__annotations__[init_arg_names[1]].__name__
        return f"{self.__class__.__name__}({self.config}, {dir_type})"

    def __getitem__(self, key):
        """Get parameter value for a given key."""
        if key in self.keys:
            if self.parameter.raw_values.dim() == 0:
                return self.parameter.raw_values
            return self.parameter.raw_values[self.keys.index(key)]
        else:
            raise ValueError(key)

    def __len__(self):
        """Return the length of raw_values."""
        return len(self.raw_values)

    @property
    def raw_values(self) -> torch.Tensor:
        """Get raw parameter values."""
        return self.parameter.raw_values

    @property
    def semantic_values(self) -> torch.Tensor:
        """Get semantic parameter values."""
        return self.parameter.semantic_values

    @property
    def readers(self) -> Dict[str, torch.Tensor]:
        """Get parameter readers."""
        return self.parameter.readers

    @readers.setter
    def readers(self, value) -> None:
        """Set parameter readers."""
        self.parameter.readers = value

    def _symmetry(self):
        """Return symmetry constraints from symmetry masks for debugging."""
        keys = np.array(self.keys)
        return [keys[mask.cpu()] for mask in self.symmetry_masks]


# --- node / Cell type parameter ------------------------------------------------


class RestingPotential(Parameter):
    """Initialize resting potentials a.k.a. biases for cell types."""

    @deepcopy_config
    def __init__(self, param_config: Namespace, connectome: ConnectomeFromAvgFilters):
        nodes_dir = connectome.nodes

        nodes = pd.DataFrame({
            k: byte_to_str(nodes_dir[k][:]) for k in param_config.groupby
        })
        grouped_nodes = nodes.groupby(
            param_config.groupby, as_index=False, sort=False
        ).first()

        param_config["type"] = grouped_nodes["type"].values
        param_config["mean"] = np.repeat(param_config["mean"], len(grouped_nodes))
        param_config["std"] = np.repeat(param_config["std"], len(grouped_nodes))

        self.parameter = forward_subclass(
            InitialDistribution, param_config, subclass_key="initial_dist"
        )
        self.indices = get_scatter_indices(nodes, grouped_nodes, param_config.groupby)
        self.keys = param_config["type"].tolist()
        self.symmetry_masks = symmetry_masks(param_config.get("symmetric", []), self.keys)


class TimeConstant(Parameter):
    """Initialize time constants for cell types."""

    @deepcopy_config
    def __init__(self, param_config: Namespace, connectome: ConnectomeFromAvgFilters):
        nodes_dir = connectome.nodes

        nodes = pd.DataFrame({
            k: byte_to_str(nodes_dir[k][:]) for k in param_config.groupby
        })
        grouped_nodes = nodes.groupby(
            param_config.groupby, as_index=False, sort=False
        ).first()

        param_config["type"] = grouped_nodes["type"].values
        param_config["value"] = np.repeat(param_config["value"], len(grouped_nodes))

        self.indices = get_scatter_indices(nodes, grouped_nodes, param_config.groupby)
        self.parameter = forward_subclass(
            InitialDistribution, param_config, subclass_key="initial_dist"
        )
        self.keys = param_config["type"].tolist()
        self.symmetry_masks = symmetry_masks(param_config.get("symmetric", []), self.keys)


# -- edge / synapse type parameter ----------------------------------------------


class SynapseSign(Parameter):
    """Initialize synapse signs for edge types."""

    @deepcopy_config
    def __init__(
        self, param_config: Namespace, connectome: ConnectomeFromAvgFilters
    ) -> None:
        edges_dir = connectome.edges

        edges = pd.DataFrame({
            k: byte_to_str(edges_dir[k][:]) for k in [*param_config.groupby, "sign"]
        })
        grouped_edges = edges.groupby(
            param_config.groupby, as_index=False, sort=False
        ).first()

        param_config.source_type = grouped_edges.source_type.values
        param_config.target_type = grouped_edges.target_type.values
        param_config.value = grouped_edges.sign.values

        self.indices = get_scatter_indices(edges, grouped_edges, param_config.groupby)
        self.parameter = forward_subclass(
            InitialDistribution, param_config, subclass_key="initial_dist"
        )
        self.keys = list(
            zip(
                param_config.source_type.tolist(),
                param_config.target_type.tolist(),
            )
        )
        self.symmetry_masks = symmetry_masks(param_config.get("symmetric", []), self.keys)


class SynapseCount(Parameter):
    """Initialize synapse counts for edge types."""

    @deepcopy_config
    def __init__(
        self, param_config: Namespace, connectome: ConnectomeFromAvgFilters
    ) -> None:
        mode = param_config.get("mode", "")
        if mode != "mean":
            raise NotImplementedError(
                f"SynapseCount does not implement {mode}. Implement "
                "a custom Parameter subclass."
            )

        edges_dir = connectome.edges

        edges = pd.DataFrame({
            k: byte_to_str(edges_dir[k][:]) for k in [*param_config.groupby, "n_syn"]
        })
        grouped_edges = edges.groupby(
            param_config.groupby, as_index=False, sort=False
        ).mean()

        param_config.source_type = grouped_edges.source_type.values
        param_config.target_type = grouped_edges.target_type.values
        param_config.du = grouped_edges.du.values
        param_config.dv = grouped_edges.dv.values

        param_config.mode = "mean"
        param_config.mean = np.log(grouped_edges.n_syn.values)

        self.indices = get_scatter_indices(edges, grouped_edges, param_config.groupby)
        self.parameter = forward_subclass(
            InitialDistribution, param_config, subclass_key="initial_dist"
        )
        self.keys = list(
            zip(
                param_config.source_type.tolist(),
                param_config.target_type.tolist(),
                param_config.du.tolist(),
                param_config.dv.tolist(),
            )
        )
        self.symmetry_masks = symmetry_masks(param_config.get("symmetric", []), self.keys)


class SynapseCountScaling(Parameter):
    """Initialize synapse count scaling for edge types.

    This class initializes synapse strengths based on the average synapse count for each
    edge type, scaling them differently for chemical and electrical synapses.

    The initialization follows this equation:

    $$\\alpha_{t_it_j} =\\frac{\\rho}{\\langle N \\rangle_{t_it_j}}$$

    where:

    1. $\\alpha_{t_it_j}$ is the synapse strength between neurons $i$ and $j$.
    2. $\\langle N \\rangle_{t_it_j}$ is the average synapse count for the edge type
        across columnar offsets $u_i-u_j$ and $v_i-v_j$
    3. $\\rho$ is a scaling factor (default: 0.01)

    """

    @deepcopy_config
    def __init__(
        self, param_config: Namespace, connectome: ConnectomeFromAvgFilters
    ) -> None:
        edges_dir = connectome.edges

        edges = pd.DataFrame({
            k: byte_to_str(edges_dir[k][:]) for k in [*param_config.groupby, "n_syn"]
        })
        grouped_edges = edges.groupby(
            param_config.groupby, as_index=False, sort=False
        ).mean()

        # to initialize synapse strengths with scale/<N>_rf
        syn_strength = param_config.get("scale", 0.01) / grouped_edges.n_syn.values

        param_config.target_type = grouped_edges.target_type.values
        param_config.source_type = grouped_edges.source_type.values
        param_config.value = syn_strength

        self.indices = get_scatter_indices(edges, grouped_edges, param_config.groupby)
        self.parameter = forward_subclass(
            InitialDistribution, param_config, subclass_key="initial_dist"
        )
        self.keys = list(
            zip(
                param_config.source_type.tolist(),
                param_config.target_type.tolist(),
            )
        )
        self.symmetry_masks = symmetry_masks(param_config.get("symmetric", []), self.keys)


def get_scatter_indices(
    dataframe: pd.DataFrame, grouped_dataframe: pd.DataFrame, groupby: List[str]
) -> Tensor:
    """Get indices for scattering operations to share parameters.

    Maps each node/edge from the complete computational graph to a parameter index.

    Args:
        dataframe: Dataframe of nodes or edges of the graph.
        grouped_dataframe: Aggregated version of the same dataframe.
        groupby: The same columns from which the grouped_dataframe was constructed.

    Returns:
        Tensor of indices for scattering operations.

    Note:
        For N elements that are grouped into M groups, this function returns N indices
        from 0 to M-1 that can be used to scatter the parameters of the M groups to the
        N elements.

    Example:
        ```python
        elements = ["A", "A", "A", "B", "B", "C", "D", "D", "E"]
        groups = ["A", "B", "C", "D", "E"]
        parameter = [1, 2, 3, 4, 5]
        # get_scatter_indices would return
        scatter_indices = [0, 0, 0, 1, 1, 2, 3, 3, 4]
        scattered_parameters = [parameter[idx] for idx in scatter_indices]
        scattered_parameters == [1, 1, 1, 2, 2, 3, 4, 4, 5]
        ```
    """
    ungrouped_elements = zip(*[dataframe[k][:] for k in groupby])
    grouped_elements = zip(*[grouped_dataframe[k][:] for k in groupby])
    to_index = {k: i for i, k in enumerate(grouped_elements)}
    return torch.tensor([to_index[k] for k in ungrouped_elements])


def symmetry_masks(
    symmetric: List[Any], keys: List[Any], as_mask: bool = False
) -> List[torch.Tensor]:
    """Create masks for subsets of parameters for joint constraints.

    Args:
        symmetric: Contains subsets of keys that point to the subsets of parameters
            to be indexed.
        keys: List of keys that point to individual parameter values.
        as_mask: If True, returns a boolean mask, otherwise integer indices.

    Returns:
        List of masks (List[torch.BoolTensor]).

    Note:
        This is experimental for configuration-based fine-grained shared parameter
        optimization, e.g. for models including multi-compartment cells or gap
        junctions.

    Example:
        ```python
        # For node type parameters with individual node types as keys:
        symmetric = [["T4a", "T4b", "T4c", "T4d"], ["T5a", "T5b", "T5c", "T5d"]]
        # This would constrain the parameter values of all T4 subtypes to their joint
        # mean and the parameter values of all T5 subtypes to their joint mean.

        # For edge type parameters with individual edge types as keys:
        symmetric = [[("CT1(M10)", "CT1(Lo1)"), ("CT1(Lo1)", "CT1(M10)")]]
        # This would constrain the edge parameter of the directed edge from CT1(M10) to
        # CT1(Lo1) and the directed edge from CT1(Lo1) to CT1(M10) to their joint mean.
        ```
    """
    if not symmetric:
        return []
    symmetry_masks = []  # type: List[torch.Tensor]
    keys = atleast_column_vector(keys)
    for identifiers in symmetric:
        identifiers = atleast_column_vector(identifiers)
        # to allow identifiers like [None, "A", None, 0]
        # for parameters that have tuples as keys
        columns = np.arange(identifiers.shape[1] + 1)[
            np.where((identifiers is not None).all(axis=0))
        ]
        try:
            symmetry_masks.append(
                torch.tensor(
                    where_equal_rows(
                        identifiers[:, columns], keys[:, columns], as_mask=as_mask
                    )
                )
            )
        except Exception as e:
            raise ValueError(
                f"{identifiers} cannot be a symmetry constraint "
                f"for parameter with keys {keys}: {e}"
            ) from e
    return symmetry_masks


# --- Supplementary Error types -------------------------------------------------


class ParameterConfigError(ValueError):
    pass



"""Classes defining the voltage initialization, voltage and current dynamics."""

from typing import Callable, Dict

import torch
from torch import nn

from flyvis.utils.tensor_utils import AutoDeref, RefTensor

__all__ = ["NetworkDynamics", "PPNeuronIGRSynapses"]

activation_fns: Dict[str, nn.Module] = {
    "relu": nn.ReLU,
    "elu": nn.ELU,
    "softplus": nn.Softplus,
    "leakyrelu": nn.LeakyReLU,
    "sigmoid": nn.Sigmoid,
    "tanh": nn.Tanh,
}

# -- The `NetworkDynamics` interface -------------------------------------------


class NetworkDynamics:
    """
    Defines the initialization and behavior of a Network during simulation.

    This class serves as an extension point for implementing custom network dynamics
    models. Subclasses must implement the following methods:

    - write_derived_params
    - write_initial_state
    - write_state_velocity

    Attributes:
        activation (nn.Module): The activation function for the network.

    Args:
        activation (dict): A dictionary specifying the activation function type and
            its parameters.
    """

    def __init__(self, activation: Dict[str, str] = {"type": "relu"}):
        self.activation = activation_fns[activation.pop("type")](**activation)

    def write_derived_params(
        self, params: AutoDeref[str, AutoDeref[str, RefTensor]], **kwargs
    ) -> None:
        """
        Augment `params`, called once at every forward pass.

        Args:
            params: A directory containing two subdirectories: `nodes` and
                `edges`, containing node and edges parameters, respectively.
            **kwargs: Additional keyword arguments.

        Note:
            This is called once per forward pass at the beginning. It's required
            after parameters have been updated by an optimizer but not at every
            timestep. Called by Network._param_api.
        """
        pass

    def write_initial_state(
        self,
        state: AutoDeref[str, AutoDeref[str, RefTensor]],
        params: AutoDeref[str, AutoDeref[str, RefTensor]],
        **kwargs,
    ) -> None:
        """
        Initialize a network's state variables from its network parameters.

        Args:
            state: A directory containing two subdirectories: `nodes` and
                `edges`. Write initial node and edge state variable values, as
                1D tensors, into them, respectively.
            params: A directory containing four subdirectories: `nodes`,
                `edges`, `sources`, and `targets`. `nodes` and `edges` contain
                node and edges parameters, respectively. `sources` and
                `targets` provide access to the node parameters associated with
                the source node and target node of each edge, respectively.
            **kwargs: Additional keyword arguments.

        Note:
            Called by Network._initial_state.
        """
        pass

    def write_state_velocity(
        self,
        vel: AutoDeref[str, AutoDeref[str, RefTensor]],
        state: AutoDeref[str, AutoDeref[str, RefTensor]],
        params: AutoDeref[str, AutoDeref[str, RefTensor]],
        target_sum: Callable,
        **kwargs,
    ) -> None:
        """
        Compute dx/dt for each state variable.

        Args:
            vel: A directory containing two subdirectories: `nodes` and
                `edges`. Write dx/dt for node and edge state variables
                into them, respectively.
            state: A directory containing two subdirectories: `nodes` and
                `edges`, containing node and edge state variable values,
                respectively.
            params: A directory containing four subdirectories: `nodes`,
                `edges`, `sources`, and `targets`. `nodes` and `edges` contain
                node and edges parameters, respectively. `sources` and
                `targets` provide access to the node parameters associated with
                the source node and target node of each edge, respectively.
            target_sum: Sums the entries in a `len(edges)` tensor corresponding
                to edges with the same target node, yielding a `len(nodes)`
                tensor.
            **kwargs: Additional keyword arguments.

        Note:
            Called by Network._next_state.
        """
        pass

    def currents(
        self,
        state: AutoDeref[str, AutoDeref[str, RefTensor]],
        params: AutoDeref[str, AutoDeref[str, RefTensor]],
    ) -> torch.Tensor:
        """
        Compute the current flowing through each edge.

        Args:
            state: A directory containing two subdirectories: `nodes` and
                `edges`, containing node and edge state variable values,
                respectively.
            params: A directory containing four subdirectories: `nodes`,
                `edges`, `sources`, and `targets`. `nodes` and `edges` contain
                node and edges parameters, respectively. `sources` and
                `targets` provide access to the node parameters associated with
                the source node and target node of each edge, respectively.

        Returns:
            A tensor of currents flowing through each edge.

        Note:
            Called by Network.current_response.
        """
        pass


# -- `NetworkDynamics` implementations -----------------------------------------


class PPNeuronIGRSynapses(NetworkDynamics):
    """Passive point neurons with instantaneous graded release synapses."""

    def write_derived_params(
        self, params: AutoDeref[str, AutoDeref[str, RefTensor]], **kwargs
    ) -> None:
        """
        Calculate weights as the product of sign, synapse count, and strength.

        Args:
            params: A directory containing edge parameters.
            **kwargs: Additional keyword arguments.
        """
        params.edges.weight = (
            params.edges.sign * params.edges.syn_count * params.edges.syn_strength
        )

    def write_initial_state(
        self,
        state: AutoDeref[str, AutoDeref[str, RefTensor]],
        params: AutoDeref[str, AutoDeref[str, RefTensor]],
        **kwargs,
    ) -> None:
        """
        Set the initial state to the bias.

        Args:
            state: A directory to write the initial state.
            params: A directory containing node parameters.
            **kwargs: Additional keyword arguments.
        """
        state.nodes.activity = params.nodes.bias

    def write_state_velocity(
        self,
        vel: AutoDeref[str, AutoDeref[str, RefTensor]],
        state: AutoDeref[str, AutoDeref[str, RefTensor]],
        params: AutoDeref[str, AutoDeref[str, RefTensor]],
        target_sum: Callable,
        x_t: torch.Tensor,
        dt: float,
        **kwargs,
    ) -> None:
        """
        Calculate velocity as bias plus sum of weighted rectified inputs.

        Args:
            vel: A directory to write the calculated velocity.
            state: A directory containing current state values.
            params: A directory containing node and edge parameters.
            target_sum: Function to sum edge values for each target node.
            x_t: External input at time t.
            dt: Time step.
            **kwargs: Additional keyword arguments.
        """
        vel.nodes.activity = (
            1
            / torch.max(params.nodes.time_const, torch.tensor(dt).float())
            * (
                -state.nodes.activity
                + params.nodes.bias
                + target_sum(
                    params.edges.weight * self.activation(state.sources.activity)
                )  # internal chemical current
                + x_t
            )
        )

    def currents(
        self,
        state: AutoDeref[str, AutoDeref[str, RefTensor]],
        params: AutoDeref[str, AutoDeref[str, RefTensor]],
    ) -> torch.Tensor:
        """
        Calculate the internal chemical current.

        Args:
            state: A directory containing current state values.
            params: A directory containing edge parameters.

        Returns:
            torch.Tensor: The calculated internal chemical current.
        """
        return params.edges.weight * self.activation(state.sources.activity)



"""Animations."""

from .activations import *
from .animations import *
from .hexflow import *
from .hexscatter import *
from .imshow import *
from .traces import *
from .sintel import *
from .network import *



"""Color coded flow-field animation."""

from typing import List, Literal, Optional, Tuple, Union

import numpy as np
import torch
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from matplotlib.text import Text

from flyvis import utils

from ..visualization import plots, plt_utils
from .animations import Animation

__all__ = ["HexFlow"]


class HexFlow(Animation):
    """Hexscatter of a color encoded flow field.

    Args:
        flow: Optic flow of shape (n_samples, n_frames, 2, n_input_elements).
        fig: Existing Figure instance or None.
        ax: Existing Axis instance or None.
        batch_sample: Batch sample to start from.
        cmap: Colormap for the hex-scatter.
        cwheel: Display colorwheel.
        cwheelxy: Colorwheel offset x and y.
        label: Label of the animation.
        labelxy: Normalized x and y location of the label.
        update: Whether to update the canvas after an animation step.
        path: Path to save the animation to.
        figsize: Figure size.
        fontsize: Font size.
        background_color: Background color of the figure and axis.

    Attributes:
        fig (Figure): Figure instance.
        ax (Axes): Axis instance.
        background_color (str): Background color of the figure and axis.
        batch_sample (int): Batch sample to start from.
        kwargs (dict): Additional keyword arguments.
        update (bool): Whether to update the canvas after an animation step.
            Must be False if this animation is composed with others using
            AnimationCollector.
        cmap: Colormap for the hex-scatter.
        cwheel (bool): Display colorwheel.
        cwheelxy (Tuple[float, float]): Colorwheel offset x and y.
        label (str): Label of the animation.
        labelxy (Tuple[float, float]): Normalized x and y location of the label.
        label_text (Text): Text instance for the label.
        sm (ScalarMappable): ScalarMappable instance for color mapping.
        fontsize (float): Font size.
        figsize (List[float, float]): Figure size.
        flow (np.ndarray): Optic flow data.
        n_samples (int): Number of samples in the flow data.
        frames (int): Number of frames in the flow data.
        extent (Tuple[float, float, float, float]): Extent of the hexagonal grid.

    Note:
        All kwargs are passed to ~flyvis.analysis.visualization.plots.hex_flow.
    """

    def __init__(
        self,
        flow: Union[np.ndarray, "torch.Tensor"],
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        batch_sample: int = 0,
        cmap: "Colormap" = plt_utils.cm_uniform_2d,
        cwheel: bool = False,
        cwheelxy: Tuple[float, float] = (),
        label: str = "Sample: {}\nFrame: {}",
        labelxy: Tuple[float, float] = (0, 1),
        update: bool = False,
        path: Optional[str] = None,
        figsize: List[float] = [2, 2],
        fontsize: float = 5,
        background_color: Literal["none"] = "none",
        **kwargs,
    ):
        self.fig = fig
        self.ax = ax
        self.background_color = background_color
        self.batch_sample = batch_sample
        self.kwargs = kwargs
        self.update = update
        self.cmap = cmap
        self.cwheel = cwheel
        self.cwheelxy = cwheelxy

        self.label = label
        self.labelxy = labelxy
        self.label_text: Optional[Text] = None
        self.sm: Optional[ScalarMappable] = None
        self.fontsize = fontsize
        self.figsize = figsize

        self.flow = utils.tensor_utils.to_numpy(flow)

        self.n_samples, self.frames = self.flow.shape[0:2]
        self.extent = utils.hex_utils.get_hextent(self.flow.shape[-1])
        super().__init__(path, self.fig)

    def init(self, frame: int = 0) -> None:
        """Initialize the animation.

        Args:
            frame: Frame number to initialize with.
        """
        u, v = utils.hex_utils.get_hex_coords(self.extent)
        self.fig, self.ax, (self.label_text, self.sm, _, _) = plots.hex_flow(
            u,
            v,
            self.flow[self.batch_sample, frame],
            fig=self.fig,
            ax=self.ax,
            cwheel=self.cwheel,
            cwheelxy=self.cwheelxy,
            cmap=self.cmap,
            annotate=False,
            labelxy=self.labelxy,
            label=self.label.format(self.batch_sample, frame),
            figsize=self.figsize,
            fontsize=self.fontsize,
            **self.kwargs,
        )
        self.fig.patch.set_facecolor(self.background_color)
        self.ax.patch.set_facecolor(self.background_color)

    def animate(self, frame: int) -> None:
        """Animate a single frame.

        Args:
            frame: Frame number to animate.
        """
        flow = self.flow[self.batch_sample, frame]

        r = np.sqrt(flow[0] ** 2 + flow[1] ** 2)
        r /= r.max()
        theta = np.arctan2(flow[1], flow[0])
        color = self.sm.to_rgba(theta)
        color[:, -1] = r

        for i, fc in enumerate(color):
            self.ax.patches[i].set_color(fc)

        if self.label:
            self.label_text.set_text(self.label.format(self.batch_sample, frame))

        if self.update:
            self.update_figure()



"""Animations of neural activations."""

from typing import List, Optional, Union

import numpy as np
from matplotlib import colormaps as cm

from flyvis import utils

from ..visualization import figsize_utils
from .animations import AnimationCollector
from .hexscatter import HexScatter

__all__ = ["StimulusResponse"]


class StimulusResponse(AnimationCollector):
    """Hex-scatter animations for input and responses.

    Args:
        stimulus: Hexagonal input.
        responses: Hexagonal activation of particular neuron type.
        batch_sample: Batch sample to start from.
        figsize_scale: Scale factor for figure size.
        fontsize: Font size for the plot.
        u: List of u coordinates of neurons to plot.
        v: List of v coordinates of neurons to plot.
        max_figure_height_cm: Maximum figure height in centimeters.
        panel_height_cm: Height of each panel in centimeters.
        max_figure_width_cm: Maximum figure width in centimeters.
        panel_width_cm: Width of each panel in centimeters.

    Attributes:
        stimulus (np.ndarray): Numpy array of stimulus data.
        responses (List[np.ndarray]): List of numpy arrays of response data.
        update (bool): Flag to indicate if update is needed.
        n_samples (int): Number of samples.
        frames (int): Number of frames.
        fig (matplotlib.figure.Figure): Matplotlib figure object.
        axes (List[matplotlib.axes.Axes]): List of matplotlib axes objects.
        animations (List[HexScatter]): List of HexScatter animation objects.
        batch_sample (int): Batch sample index.

    Note:
        If u and v are not specified, all neurons are plotted.
    """

    def __init__(
        self,
        stimulus: np.ndarray,
        responses: Union[np.ndarray, List[np.ndarray]],
        batch_sample: int = 0,
        figsize_scale: float = 1,
        fontsize: int = 5,
        u: Optional[List[int]] = None,
        v: Optional[List[int]] = None,
        max_figure_height_cm: float = 22,
        panel_height_cm: float = 3,
        max_figure_width_cm: float = 18,
        panel_width_cm: float = 3.6,
    ) -> None:
        self.stimulus = utils.tensor_utils.to_numpy(stimulus)

        # case: multiple response
        if isinstance(responses, List):
            self.responses = [utils.tensor_utils.to_numpy(r) for r in responses]
        else:
            self.responses = [utils.tensor_utils.to_numpy(responses)]

        self.update = False
        self.n_samples, self.frames = self.responses[0].shape[:2]

        figsize = figsize_utils.figsize_from_n_items(
            1 + len(self.responses),
            max_figure_height_cm=max_figure_height_cm,
            panel_height_cm=panel_height_cm,
            max_figure_width_cm=max_figure_width_cm,
            panel_width_cm=panel_width_cm,
        )
        self.fig, self.axes = figsize.axis_grid(
            unmask_n=1 + len(self.responses), hspace=0.0, wspace=0, fontsize=fontsize
        )

        stimulus_samples = self.stimulus.shape[0]
        if stimulus_samples != self.n_samples and stimulus_samples == 1:
            self.stimulus = np.repeat(self.stimulus, self.n_samples, axis=0)

        animations = []

        animations.append(
            HexScatter(
                self.stimulus,
                fig=self.fig,
                ax=self.axes[0],
                title="stimulus",
                labelxy=(-0.1, 1),
                update=False,
                title_y=0.9,
            )
        )

        cranges = np.max(np.abs(self.responses), axis=(0, 2, 3, 4))

        for i, responses in enumerate(self.responses, 1):
            animations.append(
                HexScatter(
                    responses,
                    fig=self.fig,
                    ax=self.axes[i],
                    cmap=cm.get_cmap("seismic"),
                    title=f"response {i}" if len(self.responses) > 1 else "response",
                    label="",
                    midpoint=0,
                    update=False,
                    u=u,
                    v=v,
                    cranges=cranges,
                    cbar=i == len(self.responses),
                    title_y=0.9,
                )
            )

        self.animations = animations
        self.batch_sample = batch_sample
        super().__init__(None, self.fig)



"""Plotting functions."""

# -- Set default matplotlib configuration
from importlib.resources import files

import matplotlib
import matplotlib.pyplot as plt

matplotlibrc_path = files("flyvis.analysis.visualization").joinpath("matplotlibrc")
matplotlib.rc_file(matplotlibrc_path)

plt.set_loglevel(level="warning")
del matplotlib, plt



from unittest.mock import patch

import pytest

from flyvis_cli.flyvis_cli import (
    SCRIPT_COMMANDS,
    filter_args,
    handle_help_request,
    main,
)


def test_filter_args():
    # Test filtering out commands from arguments
    argv = [
        "--ensemble_id",
        "0001",
        "train",
        "--task_name",
        "flow",
        "validate",
        "task.n_iters=100",
        "foo:int=1",
        "record",
    ]
    commands = ["train", "validate", "record"]
    selected_commands, filtered_args = filter_args(argv, commands)
    assert selected_commands == ["train", "validate", "record"]
    assert filtered_args == [
        "--ensemble_id",
        "0001",
        "--task_name",
        "flow",
        "task.n_iters=100",
        "foo:int=1",
    ]

    # Test with no commands to filter
    argv = ["--ensemble_id", "0001", "--task_name", "flow"]
    commands = ["train", "validate"]
    selected_commands, filtered_args = filter_args(argv, commands)
    assert selected_commands == []
    assert filtered_args == argv

    # Test with empty input
    selected_commands, filtered_args = filter_args([], [])
    assert selected_commands == []
    assert filtered_args == []

    # Test order preservation
    argv = ["validate", "--ensemble_id", "0001", "train", "--task_name", "flow"]
    commands = ["train", "validate"]
    selected_commands, filtered_args = filter_args(argv, commands)
    assert selected_commands == ["validate", "train"]  # Commands preserved in order found
    assert filtered_args == ["--ensemble_id", "0001", "--task_name", "flow"]


def test_handle_help_request():
    # Test help request for a specific command
    with patch('flyvis_cli.flyvis_cli.run_script') as mock_run:
        argv = ["flyvis", "train", "--help"]
        assert handle_help_request(argv) is True
        mock_run.assert_called_once_with(SCRIPT_COMMANDS["train"], ["--help"])

    # Test no help request
    with patch('flyvis_cli.flyvis_cli.run_script') as mock_run:
        argv = ["flyvis", "train", "--ensemble_id", "0001"]
        assert handle_help_request(argv) is False
        mock_run.assert_not_called()


@pytest.mark.parametrize(
    "args,expected_behavior",
    [
        (
            ["--ensemble_id", "0001", "--task_name", "flow", "train"],
            {"should_succeed": True, "expected_command": "train"},
        ),
        (
            ["--help"],
            {"should_succeed": False, "return_code": 1},
        ),
        (
            [],
            {"should_succeed": False, "return_code": 1},
        ),
    ],
)
def test_main_argument_parsing(args, expected_behavior):
    """Test CLI argument parsing behavior."""
    with (
        patch('sys.argv', ['flyvis'] + args),
        patch('flyvis_cli.flyvis_cli.run_script') as mock_run,
    ):
        result = main()

        if expected_behavior["should_succeed"]:
            assert result == 0
            # Verify the correct script was called
            mock_run.assert_called_once()
            script_path = mock_run.call_args[0][0]
            assert script_path.name == f"{expected_behavior['expected_command']}.py"
        else:
            # Help and usage errors return 1
            assert result == expected_behavior["return_code"]


def test_main_runs_multiple_commands():
    args = ["--ensemble_id", "0001", "--task_name", "flow", "train", "validate"]
    with (
        patch('sys.argv', ['flyvis'] + args),
        patch('flyvis_cli.flyvis_cli.run_script') as mock_run,
    ):
        main()
        assert mock_run.call_count == 2
        # Verify calls were made in the correct order
        expected_calls = [
            (
                (
                    SCRIPT_COMMANDS["train"],
                    ["--ensemble_id", "0001", "--task_name", "flow"],
                ),
            ),
            (
                (
                    SCRIPT_COMMANDS["validate"],
                    ["--ensemble_id", "0001", "--task_name", "flow"],
                ),
            ),
        ]
        mock_run.assert_has_calls(expected_calls)



"""Transformations and augmentations on hex-lattices."""

from typing import Any, List, Optional

import numpy as np
import torch

from .augmentation import Augmentation
from .utils import (
    flip_matrix,
    flip_permutation_index,
    rotation_matrix,
    rotation_permutation_index,
)

__all__ = [
    "HexRotate",
    "HexFlip",
    "ContrastBrightness",
    "PixelNoise",
    "GammaCorrection",
]


class HexRotate(Augmentation):
    """Rotate a sequence of regular hex-lattices by multiple of 60 degree.

    Args:
        extent: Extent of the regular hexagonal grid in columns.
        n_rot: Number of 60 degree rotations. 0-5.
        p_rot: Probability of rotating. If None, no rotation is performed.

    Attributes:
        extent (int): Extent of the regular hexagonal grid in columns.
        n_rot (int): Number of 60 degree rotations. 0-5.
        p_rot (float): Probability of rotating.
        permutation_indices (dict): Cached indices for rotation.
        rotation_matrices (dict): Cached rotation matrices for rotation.
    """

    def __init__(self, extent: int, n_rot: int = 0, p_rot: float = 0.5) -> None:
        self.extent = extent
        self.rotation_matrices: dict = {}
        self.permutation_indices: dict = {}
        for n in range(6):
            R2 = rotation_matrix(n * 60 * np.pi / 180, three_d=False)
            R3 = rotation_matrix(n * 60 * np.pi / 180, three_d=True)
            self.rotation_matrices[n] = [R2, R3]
            self.permutation_indices[n] = rotation_permutation_index(extent, n)
        self.n_rot = n_rot
        self.p_rot = p_rot
        self.set_or_sample(n_rot)

    @property
    def n_rot(self) -> int:
        """Get the number of rotations."""
        return self._n_rot

    @n_rot.setter
    def n_rot(self, n_rot: int) -> None:
        """Set the number of rotations."""
        self._n_rot = n_rot % 6

    def rotate(self, seq: torch.Tensor) -> torch.Tensor:
        """Rotate a sequence on a regular hexagonal lattice.

        Args:
            seq: Sequence of shape (frames, dims, hexals).

        Returns:
            Rotated sequence of the same shape as seq.
        """
        dims = seq.shape[-2]
        seq = seq[..., self.permutation_indices[self.n_rot]]
        if dims > 1:
            seq = self.rotation_matrices[self.n_rot][dims - 2] @ seq
        return seq

    def transform(self, seq: torch.Tensor, n_rot: Optional[int] = None) -> torch.Tensor:
        """Rotate a sequence on a regular hexagonal lattice.

        Args:
            seq: Sequence of shape (frames, dims, hexals).
            n_rot: Optional number of rotations to apply.

        Returns:
            Rotated sequence of the same shape as seq.
        """
        if n_rot is not None:
            self.n_rot = n_rot
        if self.n_rot > 0:
            return self.rotate(seq)
        return seq

    def set_or_sample(self, n_rot: Optional[int] = None) -> None:
        """Set or sample the number of rotations.

        Args:
            n_rot: Number of rotations to set. If None, sample randomly.
        """
        if n_rot is None:
            n_rot = (
                np.random.randint(low=1, high=6)
                if self.p_rot and self.p_rot > np.random.rand()
                else 0
            )
        self.n_rot = n_rot


class HexFlip(Augmentation):
    """Flip a sequence of regular hex-lattices across one of three hex-axes.

    Args:
        extent: Extent of the regular hexagonal grid.
        axis: Flipping axis. 0 corresponds to no flipping.
        p_flip: Probability of flipping. If None, no flipping is performed.
        flip_axes: List of valid flipping axes. Can contain 0, 1, 2, 3.

    Attributes:
        extent (int): Extent of the regular hexagonal grid.
        axis (int): Flipping axis.
        p_flip (float): Probability of flipping.
        flip_axes (List[int]): List of valid flipping axes.
        permutation_indices (dict): Cached indices for flipping.
        rotation_matrices (dict): Cached rotation matrices for flipping.

    Note:
        This is to avoid redundant transformations from rotation and flipping.
        For example, flipping across the 1st axis is equivalent to rotating by
        240 degrees and flipping across the 2nd axis.
    """

    def __init__(
        self,
        extent: int,
        axis: int = 0,
        p_flip: float = 0.5,
        flip_axes: List[int] = [0, 1, 2, 3],
    ) -> None:
        self.extent = extent
        self.rotation_matrices: dict = {}
        self.permutation_indices: dict = {}
        for n, angle in enumerate([90, 150, 210], 1):
            R2 = flip_matrix(np.radians(angle), three_d=False)
            R3 = flip_matrix(np.radians(angle), three_d=True)
            self.rotation_matrices[n] = [R2, R3]
            self.permutation_indices[n] = flip_permutation_index(extent, n)
        self.flip_axes = flip_axes
        self.axis = axis
        self.p_flip = p_flip
        self.set_or_sample(axis)

    @property
    def axis(self) -> int:
        """Get the flipping axis."""
        return self._axis

    @axis.setter
    def axis(self, axis: int) -> None:
        """Set the flipping axis."""
        assert (
            axis in self.flip_axes
        ), f"{axis} is not a valid axis. Must be in {self.flip_axes}."
        self._axis = axis

    def flip(self, seq: torch.Tensor) -> torch.Tensor:
        """Flip a sequence on a regular hexagonal lattice.

        Args:
            seq: Sequence of shape (frames, dims, hexals).

        Returns:
            Flipped sequence of the same shape as seq.
        """
        dims = seq.shape[-2]
        seq = seq[..., self.permutation_indices[self.axis]]
        if dims > 1:
            seq = self.rotation_matrices[self.axis][dims - 2] @ seq
        return seq

    def transform(self, seq: torch.Tensor, axis: Optional[int] = None) -> torch.Tensor:
        """Flip a sequence on a regular hexagonal lattice.

        Args:
            seq: Sequence of shape (frames, dims, hexals).
            axis: Optional flipping axis to apply.

        Returns:
            Flipped sequence of the same shape as seq.
        """
        if axis is not None:
            self.axis = axis
        if self.axis > 0:
            return self.flip(seq=seq)
        return seq

    def set_or_sample(self, axis: Optional[int] = None) -> None:
        """Set or sample the flipping axis.

        Args:
            axis: Flipping axis to set. If None, sample randomly.
        """
        if axis is None:
            axis = (
                np.random.randint(low=1, high=max(self.flip_axes) + 1)
                if self.p_flip and self.p_flip > np.random.rand()
                else 0
            )
        self.axis = axis


class ContrastBrightness(Augmentation):
    """Contrast transformation.

    The transformation is described as:
    ```python
    pixel = max(0, contrast_factor * (pixel - 0.5) + 0.5
                + contrast_factor * brightness_factor)
    ```

    Args:
        contrast_factor: Contrast factor.
        brightness_factor: Brightness factor.
        contrast_std: Standard deviation of the contrast factor.
        brightness_std: Standard deviation of the brightness factor.

    Attributes:
        contrast_std (float): Standard deviation of the contrast factor.
        brightness_std (float): Standard deviation of the brightness factor.
        contrast_factor (float): Contrast factor.
        brightness_factor (float): Brightness factor.
    """

    def __init__(
        self,
        contrast_factor: Optional[float] = None,
        brightness_factor: Optional[float] = None,
        contrast_std: float = 0.2,
        brightness_std: float = 0.1,
    ) -> None:
        self.contrast_std = contrast_std
        self.brightness_std = brightness_std
        self.set_or_sample(contrast_factor, brightness_factor)

    def transform(self, seq: torch.Tensor) -> torch.Tensor:
        """Apply the transformation to a sequence.

        Args:
            seq: Input sequence.

        Returns:
            Transformed sequence.
        """
        if self.contrast_factor is not None:
            return (
                self.contrast_factor * (seq - 0.5)
                + 0.5
                + self.contrast_factor * self.brightness_factor
            ).clamp(0)
        return seq

    def set_or_sample(
        self,
        contrast_factor: Optional[float] = None,
        brightness_factor: Optional[float] = None,
    ) -> None:
        """Set or sample contrast and brightness factors.

        Args:
            contrast_factor: Contrast factor to set. If None, sample randomly.
            brightness_factor: Brightness factor to set. If None, sample randomly.
        """
        if contrast_factor is None:
            # behaves like N(1, std) for small std, slightly biased towards
            # high contrast in particular for large std deviations
            # TODO: implement other sampling schemes
            contrast_factor = (
                np.exp(np.random.normal(0, self.contrast_std))
                if self.contrast_std
                else None
            )
        if brightness_factor is None:
            brightness_factor = (
                np.random.normal(0, self.brightness_std) if self.brightness_std else 0.0
            )
        self.contrast_factor = contrast_factor
        self.brightness_factor = brightness_factor


class PixelNoise(Augmentation):
    """Pixelwise gaussian noise.

    The transformation is described as:
    ```python
    pixel = pixel + N(0, std)
    ```

    It biases the signal to noise ratio: high for light, low for dark pixels.

    Args:
        std: Standard deviation of the gaussian noise.

    Attributes:
        std (float): Standard deviation of the gaussian noise.
    """

    def __init__(self, std: float = 0.08) -> None:
        self.std = std

    def transform(self, seq: torch.Tensor) -> torch.Tensor:
        """Apply the transformation to a sequence.

        Args:
            seq: Input sequence.

        Returns:
            Transformed sequence.
        """
        if self.std:
            noise = torch.randn_like(seq) * self.std
            return (seq + noise).clamp(0)
        return seq

    def set_or_sample(self, std: Optional[float] = None) -> None:
        """Set or sample the standard deviation of the gaussian noise.

        Args:
            std: Standard deviation of the gaussian noise to set.
                If None, no change is made.
        """
        if std is None:
            return
        self.std = std


class GammaCorrection(Augmentation):
    """Gamma correction.

    The transformation is described as:
    ```python
    pixel = pixel ** gamma
    ```

    Gamma > 1 increases the contrast, gamma < 1 decreases the contrast.

    Args:
        gamma: Gamma value.
        std: Standard deviation of the gamma value.

    Attributes:
        gamma: float
            Gamma value.
        std: Optional[float]
            Standard deviation of the gamma value.
    """

    def __init__(self, gamma: float = 1, std: Optional[float] = None) -> None:
        self.gamma = gamma
        self.std = std

    def transform(self, seq: torch.Tensor) -> torch.Tensor:
        """Apply the transformation to a sequence.

        Args:
            seq: Input sequence.

        Returns:
            Transformed sequence.
        """
        if self.gamma:
            return seq**self.gamma
        return seq

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "gamma" and value < 0:
            raise ValueError("Gamma must be positive.")
        return super().__setattr__(name, value)

    def set_or_sample(self, gamma: Optional[float] = None) -> None:
        """Set or sample the gamma value.

        Args:
            gamma: Gamma value to set. If None, sample randomly.
        """
        if gamma is None:
            gamma = max(0, np.random.normal(1, self.std)) if self.std else 1.0
        self.gamma = gamma



import subprocess
import sys
from pathlib import Path

# new scripts need to be added here to be included in the docs
HEADERS = {
    "download_pretrained_models.py": "Download Pretrained Models",
    "init_config.py": "Initialize Configuration",
    "flyvis_cli.py": "Command Line Interface Entry Point",
    "training/train_single.py": "Run Training for Single Model",
    "training/train.py": "Launch Ensemble Training on Compute Cloud",
    "validation/val_single.py": "Run Validation for Single Model",
    "validation/validate.py": "Launch Ensemble Validation on Compute Cloud",
    "analysis/synthetic_recordings_single.py": "Run Synthetic Recordings",
    "analysis/record.py": "Launch Synthetic Recordings on Compute Cloud",
    "analysis/ensemble_analysis.py": "Run Ensemble Analysis",
    "analysis/analysis.py": "Launch Ensemble Analysis on Compute Cloud",
    "analysis/notebook.py": "Run Notebook",
    "analysis/notebook_per_ensemble.py": "Launch Notebook Per Ensemble on Compute Cloud",
    "analysis/notebook_per_model.py": "Launch Notebook Per Model on Compute Cloud",
}


def capture_help(script_path: str) -> str:
    """Runs the script with --help and captures the output."""
    result = subprocess.run(
        [sys.executable, script_path, '--help'], capture_output=True, text=True
    )
    return result.stdout


def script_to_md(script_path: Path, output_dir: Path, scripts_dir: Path):
    """Converts a script's help output to a markdown file."""
    help_output = capture_help(str(script_path))

    # Create the output filename
    relative_path = script_path.relative_to(scripts_dir)
    output_file = output_dir / f"{relative_path.with_suffix('.md')}"

    # Ensure the output directory exists
    output_file.parent.mkdir(parents=True, exist_ok=True)

    # Generate the script path for documentation
    doc_script_path = f"""
::: flyvis_cli.{str(relative_path).replace('/', '.').replace('.py', '')}
    options:
      heading_level: 4
"""

    # Write the help message to a markdown file
    with open(output_file, "w", encoding="utf-8") as f:
        f.write(f"# {HEADERS[str(relative_path)]}\n\n")
        f.write(f"{doc_script_path}\n\n")
        f.write("```\n" + help_output + "\n```")


def main():
    # Path to the scripts directory
    scripts_dir = Path(__file__).parent.parent / "flyvis_cli"

    # Path to the output directory
    output_dir = Path(__file__).parent / "docs" / "reference" / "flyvis_cli"

    # Process all Python scripts in the scripts directory and its subdirectories
    for script_path in scripts_dir.rglob("*.py"):
        if script_path.name == "__init__.py":
            continue
        print(script_path)
        script_to_md(script_path, output_dir, scripts_dir)


if __name__ == "__main__":
    main()



"""Utility functions for operations on pandas DataFrames."""

from typing import Iterable

from pandas import DataFrame


def filter_by_column_values(
    dataframe: DataFrame, column: str, values: Iterable
) -> DataFrame:
    """
    Return subset of dataframe based on list of values to appear in a column.

    Args:
        dataframe: DataFrame with key as column.
        column: Column of the dataframe, e.g. `type`.
        values: Types of neurons e.g. R1, T4a, etc.

    Returns:
        DataFrame: Subset of the input dataframe.

    Example:
        ```python
        filtered_df = filter_by_column_values(df, 'neuron_type', ['R1', 'T4a'])
        ```
    """
    cond = ""
    for t in values:
        cond += f"(dataframe.{column}=='{t}')"
        if t != values[-1]:
            cond += "|"
    return dataframe[eval(cond)]


def where_dataframe(arg_df: DataFrame, **kwargs) -> DataFrame:
    """
    Return indices of rows in a DataFrame where conditions are met.

    Conditions are passed as keyword arguments.

    Args:
        arg_df: Input DataFrame.
        **kwargs: Keyword arguments representing conditions.

    Returns:
        DataFrame: Indices of rows where conditions are met.

    Example:
        ```python
        indices = where_dataframe(df, type='T4a', u=2, v=0)
        ```

    Note:
        The dataframe is expected to have columns matching the keyword arguments.
    """

    def _query_from_kwargs(kwargs):
        _query_start = "{}=={}"
        _query_append = "& {}=={}"

        _query_elements = []
        for i, (key, value) in enumerate(kwargs.items()):
            if isinstance(value, str) and (
                not value.startswith("'") or value.startswith('"')
            ):
                value = f"'{value}'"
            if i == 0:
                _query_elements.append(_query_start.format(key, value))
            else:
                _query_elements.append(_query_append.format(key, value))
        return "".join(_query_elements)

    query = _query_from_kwargs(kwargs)

    return arg_df.query(query).index



import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Union

import flyvis
from flyvis.network.ensemble import model_paths_from_parent


@dataclass
class Status:
    """Status object from log files of model runs.

    Attributes:
        ensemble_name: Name of the ensemble.
        log_files: List of all log files.
        train_logs: List of train logs.
        model_id_to_train_log_file: Mapping of model ID to log file.
        status: Mapping of model ID to status.
        user_input: Mapping of model ID to user input (behind LSF command).
        hosts: Mapping of model ID to host.
        rerun_failed_runs: Formatted submission commands to restart failed models.
        lsf_part: LSF command part.
    """

    ensemble_name: str
    log_files: List[Path]
    train_logs: List[Path]
    model_id_to_train_log_file: Dict[str, Path]
    status: Dict[str, str]
    user_input: Dict[str, str]
    hosts: Dict[str, List[str]]
    rerun_failed_runs: Dict[str, List[str]]
    lsf_part: str

    def print_for_rerun(
        self, exclude_failed_hosts: bool = True, model_ids: List[str] = None
    ) -> None:
        """Print formatted submission commands to restart failed models.

        Args:
            exclude_failed_hosts: Whether to exclude failed hosts.
            model_ids: List of model IDs to rerun. If None, all failed models are
                included.
        """
        model_ids = model_ids or list(self.rerun_failed_runs.keys())
        for model_id in model_ids:
            command = ""
            subcmds = self.rerun_failed_runs[model_id]
            if exclude_failed_hosts:
                command += f"{subcmds[0]}{subcmds[1]}{subcmds[2]}"
            else:
                command += f"{subcmds[0]}{subcmds[2]}"
            print(command)

    def get_hosts(self) -> List[str]:
        """Get hosts on which the job was executed."""
        return list(set(flatten_list(self.hosts.values())))

    def bad_hosts(self) -> List[str]:
        """Get hosts on which the job failed."""
        host_lists = [
            host
            for model_id, host in self.hosts.items()
            if self.status[model_id] not in ["Successfully completed.", "running"]
        ]
        return list(set(flatten_list(host_lists)))

    def successful_runs(self) -> int:
        """Get number of successful runs."""
        return sum(1 for v in self.status.values() if v == "Successfully completed.")

    def running_runs(self) -> int:
        """Get number of running runs."""
        return sum(1 for v in self.status.values() if v == "running")

    def failed_runs(self) -> int:
        """Get number of failed runs."""
        return sum(1 for v in self.status.values() if "Exited with exit code" in v)

    def successful_model_ids(self) -> List[str]:
        """Get model IDs of successful runs."""
        return [k for k, v in self.status.items() if v == "Successfully completed."]

    def running_model_ids(self) -> List[str]:
        """Get model IDs of running runs."""
        return [k for k, v in self.status.items() if v == "running"]

    def failed_model_ids(self) -> List[str]:
        """Get model IDs of failed runs."""
        return [k for k, v in self.status.items() if "Exited with exit code" in v]

    def lookup_log(
        self, model_id: str, log_type: str = "train_single", last_n_lines: int = 20
    ) -> List[str]:
        """Lookup log for a model ID.

        Args:
            model_id: ID of the model.
            log_type: Type of log to lookup.
            last_n_lines: Number of lines to return from the end of the log.

        Returns:
            List of log lines.
        """
        log_file = [
            p
            for p in self.log_files
            if log_type in str(p) and p.name.split("_")[0] == model_id
        ][0]
        return log_file.read_text().split("\n")[-last_n_lines:]

    def extract_error_trace(
        self, model_id: str, check_last_n_lines: int = 100, log_type: str = "train_single"
    ) -> str:
        """Extract the Python error message and traceback from a given log string.

        Args:
            model_id: ID of the model.
            check_last_n_lines: Number of lines to check from the end of the log.
            log_type: Type of log to extract error from.

        Returns:
            Extracted error message and traceback, or a message if no error is found.
        """
        log_string = "\n".join(
            self.lookup_log(model_id, last_n_lines=check_last_n_lines, log_type=log_type)
        )
        pattern = r"Traceback \(most recent call last\):(.+?)(?=\n\n|\Z)"
        match = re.search(pattern, log_string, re.DOTALL)
        return match.group(0).strip() if match else "No Python error found in the log."

    def extract_error_type(
        self, model_id: str, log_type: str = "train_single", check_last_n_lines: int = 100
    ) -> str:
        """Extract the type of error from a given log string.

        Args:
            model_id: ID of the model.
            log_type: Type of log to extract error from.
            check_last_n_lines: Number of lines to check from the end of the log.

        Returns:
            Extracted error type, or a message if no specific error type is found.
        """
        log_string = "\n".join(
            self.lookup_log(model_id, last_n_lines=check_last_n_lines, log_type=log_type)
        )
        pattern = r"\b[A-Z]\w*Error\b"
        match = re.search(pattern, log_string)
        return match.group(0) if match else "No specific error type found."

    def all_errors(
        self, check_last_n_lines: int = 100, log_type: str = "train_single"
    ) -> set:
        """Get all unique errors from failed runs.

        Args:
            check_last_n_lines: Number of lines to check from the end of the log.
            log_type: Type of log to extract errors from.

        Returns:
            Set of unique error messages.
        """
        return set(
            self.extract_error_trace(
                model_id,
                check_last_n_lines=check_last_n_lines,
                log_type=log_type,
            )
            for model_id in self.failed_model_ids()
        )

    def all_error_types(self, log_type: str = "train_single") -> set:
        """Get all unique error types from failed runs.

        Args:
            log_type: Type of log to extract error types from.

        Returns:
            Set of unique error types.
        """
        return set(
            self.extract_error_type(model_id, log_type=log_type)
            for model_id in self.failed_model_ids()
        )

    def print_all_errors(
        self, check_last_n_lines: int = 100, log_type: str = "train_single"
    ) -> None:
        """Print all errors and tracebacks from failed runs.

        Args:
            check_last_n_lines: Number of lines to check from the end of the log.
            log_type: Type of log to extract errors from.
        """
        for model_id in self.failed_model_ids():
            print(
                f"Model {model_id} failed with the following error message "
                "and traceback:\n"
            )
            print(
                self.extract_error_trace(
                    model_id,
                    check_last_n_lines=check_last_n_lines,
                    log_type=log_type,
                )
            )
            print("\n")

    def __getitem__(self, key: str) -> str:
        """Get status for a specific model ID."""
        if key in self.status:
            return self.status[key]
        return object.__getitem__(self, key)

    def __repr__(self) -> str:
        """Return a string representation of the Status object."""
        _repr = f"Status of ensemble {self.ensemble_name}."
        _repr += f"\n{len(self.status)} models."
        _repr += f"\nHosts: {','.join(self.get_hosts())}."
        _repr += f"\n  {self.successful_runs()} successful runs."
        _repr += f"\n  {self.running_runs()} running runs."
        _repr += f"\n  {self.failed_runs()} failed runs."
        if self.failed_runs() > 0:
            _repr += f"\n  Bad hosts: {','.join(self.bad_hosts())}."
            _repr += "\n  Use .print_for_rerun() to print formatted submission commands"
            _repr += " to restart failed models."
            _repr += "\nError types:"
            for error in self.all_error_types():
                _repr += f"\n  {error}"
            _repr += (
                "\n  Run .print_all_errors() to print the error messages and tracebacks."
            )
        return _repr


def find_host(log_string: str) -> List[str]:
    """Find the host(s) on which the job was executed.

    Args:
        log_string: The log string to search for host information.

    Returns:
        List of host names.
    """
    pattern = r"executed on host\(s\) <(?:\d*\*)?(.+?)>,"
    return re.findall(pattern, log_string)


def get_exclude_host_part(log_string: str, exclude_hosts: Union[str, List[str]]) -> str:
    """Get the part of the LSF command that excludes hosts.

    Args:
        log_string: The log string to search for host information.
        exclude_hosts: Host(s) to exclude. Can be 'auto', a single host name, or a list
            of host names.

    Returns:
        The LSF command part for excluding hosts.
    """
    if exclude_hosts is None:
        return ""

    exclude_host_part = '-R "select[{}]" '

    if isinstance(exclude_hosts, str) and exclude_hosts == "auto":
        exclude_hosts = find_host(log_string)
    elif isinstance(exclude_hosts, str):
        exclude_hosts = [exclude_hosts]

    exclusion_strings = [f"hname!='{host}'" for host in exclude_hosts]
    return exclude_host_part.format(" && ".join(exclusion_strings))


def get_status(
    ensemble_name: str,
    nP: int = 4,
    gpu: str = "num=1",
    queue: str = "gpu_l4",
    exclude_hosts: Union[str, List[str]] = "auto",
) -> Status:
    """Get Status object for the ensemble of models with formatting for rerun.

    Args:
        ensemble_name: Ensemble name (e.g. "flow/<id>").
        nP: Number of processors.
        gpu: Number of GPUs.
        queue: Queue name.
        exclude_hosts: Host(s) to exclude. Can be 'auto', a single host name, or a
            list of host names.

    Returns:
        Status object containing information about the ensemble runs.
    """
    _lsf_part = "bsub -J {} -n {} -o {} -gpu '{}' -q {} "

    tnn_paths, path = model_paths_from_parent(flyvis.results_dir / ensemble_name)
    log_files = [p for p in path.iterdir() if p.suffix == ".log"]
    train_logs = [p for p in log_files if "train_single" in str(p)]
    model_id_to_train_log_file = {p.name.split("_")[0]: p for p in train_logs}

    status = {}
    user_input = {}
    hosts = {}
    log_strings = {}
    for p in train_logs:
        model_id = p.name.split("_")[0]
        log_str = p.read_text()
        log_strings[model_id] = log_str
        if log_str.split("\n")[-3] == "The output (if any) is above this job summary.":
            status[model_id] = log_str.split("\n")[-18]
            user_input[model_id] = log_str.split("\n")[-21]
        else:
            status[model_id] = "running"
            user_input[model_id] = ""

        hosts[model_id] = find_host(log_str)

    _lfs_cmd = _lsf_part
    rerun_failed_runs = {}
    for model_id, stat in status.items():
        if stat not in ["Successfully completed.", "running"]:
            _lsf_cmd = _lsf_part.format(
                f"{ensemble_name}/{model_id}",
                nP,
                model_id_to_train_log_file[model_id],
                gpu,
                queue,
            )
            exclude_host_part = get_exclude_host_part(
                log_strings[model_id], exclude_hosts
            )
            rerun_failed_runs[model_id] = [
                _lsf_cmd,
                exclude_host_part,
                user_input[model_id],
            ]
    return Status(
        ensemble_name,
        log_files,
        train_logs,
        model_id_to_train_log_file,
        status,
        user_input,
        hosts,
        rerun_failed_runs,
        _lfs_cmd,
    )


def flatten_list(nested_list: List) -> List:
    """Flatten a nested list of lists into a single list with all elements.

    Args:
        nested_list: A nested list of lists to be flattened.

    Returns:
        A single flattened list with all elements.
    """
    flattened = []
    for item in nested_list:
        if isinstance(item, list):
            flattened.extend(flatten_list(item))
        else:
            flattened.append(item)
    return flattened



import logging
import multiprocessing
import os
import re
import signal
import socket
import subprocess
import sys
import warnings
from abc import ABC, abstractmethod
from time import sleep
from typing import Dict, List

from flyvis import results_dir

logging.basicConfig(
    format="[%(asctime)s] [%(filename)s:%(lineno)d] %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class ClusterManager(ABC):
    """Abstract base class for cluster management operations."""

    @abstractmethod
    def run_job(self, command: str) -> str:
        """
        Run a job on the cluster.

        Args:
            command: The command to run.

        Returns:
            The job ID as a string.
        """
        pass

    @abstractmethod
    def is_running(self, job_id: str) -> bool:
        """
        Check if a job is running.

        Args:
            job_id: The ID of the job to check.

        Returns:
            True if the job is running, False otherwise.
        """
        pass

    @abstractmethod
    def kill_job(self, job_id: str) -> str:
        """
        Kill a running job.

        Args:
            job_id: The ID of the job to kill.

        Returns:
            The output of the kill command.
        """
        pass

    @abstractmethod
    def get_submit_command(
        self, job_name: str, n_cpus: int, output_file: str, gpu: str, queue: str
    ) -> str:
        """
        Get the command to submit a job to the cluster.

        Args:
            job_name: The name of the job.
            n_cpus: The number of CPUs to request.
            output_file: The file to write job output to.
            gpu: The GPU configuration.
            queue: The queue to submit the job to.

        Returns:
            The submit command as a string.
        """
        pass

    @abstractmethod
    def get_script_part(self, command: str) -> str:
        """
        Get the script part of the command.

        Args:
            command: The command to wrap.

        Returns:
            The wrapped command as a string.
        """
        pass


class LSFManager(ClusterManager):
    """Cluster manager for LSF (Load Sharing Facility) systems."""

    def run_job(self, command: str) -> str:
        answer = subprocess.getoutput(command)
        job_id = re.findall(r"(?<=<)\d+(?=>)", answer)
        assert len(job_id) == 1
        return job_id[0]

    def is_running(self, job_id: str) -> bool:
        job_info = subprocess.getoutput("bjobs -w")
        return job_id in job_info

    def kill_job(self, job_id: str) -> str:
        return subprocess.getoutput(f"bkill {job_id}")

    def get_submit_command(
        self, job_name: str, n_cpus: int, output_file: str, gpu: str, queue: str
    ) -> str:
        return f"bsub -J {job_name} -n {n_cpus} -o {output_file} -gpu '{gpu}' -q {queue} "

    def get_script_part(self, command: str) -> str:
        return command


class SLURMManager(ClusterManager):
    """Cluster manager for SLURM systems.

    Warning:
        This is untested.
    """

    def run_job(self, command: str) -> str:
        answer = subprocess.getoutput(command)
        job_id = re.findall(r"\d+", answer)
        assert len(job_id) == 1
        return job_id[0]

    def is_running(self, job_id: str) -> bool:
        cmd = f"sacct -j {job_id} --format=State --noheader -X"
        state = subprocess.getoutput(cmd).strip()
        return state in ["PENDING", "RUNNING", "REQUEUED"]

    def kill_job(self, job_id: str) -> str:
        return subprocess.getoutput(f"scancel {job_id}")

    def get_submit_command(
        self, job_name: str, n_cpus: int, output_file: str, gpu: str, queue: str
    ) -> str:
        return (
            f"sbatch --job-name={job_name} "
            f"--cpus-per-task={n_cpus} "
            f"--output={output_file} "
            f"--gres=gpu:{gpu} "
            f"--partition={queue} "
        )

    def get_script_part(self, command: str) -> str:
        return f"--wrap '{command}'"


class VirtualClusterManager(ClusterManager):
    """Simulated cluster manager for local execution."""

    def __init__(self):
        warnings.warn(
            "VirtualClusterManager is being used. This is not recommended for "
            "production use and may not accurately represent cluster behavior.",
            UserWarning,
            stacklevel=2,
        )
        self.running_jobs = {}
        self._current_output_file = None

    def get_submit_command(
        self, job_name: str, n_cpus: int, output_file: str, gpu: str, queue: str
    ) -> str:
        self._current_output_file = output_file
        return f"echo 'Running {job_name} locally' && "

    def run_job(self, command: str) -> str:
        job_id = str(len(self.running_jobs) + 1)
        if self._current_output_file:
            # Create the output directory if it doesn't exist
            output_dir = os.path.dirname(self._current_output_file)
            logger.info(f"Writing output to: {self._current_output_file}")
            if output_dir:
                os.makedirs(output_dir, exist_ok=True)

            with open(self._current_output_file, 'w') as f:
                process = multiprocessing.Process(
                    target=subprocess.run,
                    args=(command,),
                    kwargs={'shell': True, 'stdout': f, 'stderr': subprocess.STDOUT},
                )
                process.start()
                self.running_jobs[job_id] = process
        else:
            logger.warning("No output file specified for job")
            process = multiprocessing.Process(target=os.system, args=(command,))
            process.start()
            self.running_jobs[job_id] = process
        return job_id

    def is_running(self, job_id: str) -> bool:
        return job_id in self.running_jobs and self.running_jobs[job_id].is_alive()

    def kill_job(self, job_id: str) -> str:
        if job_id in self.running_jobs:
            process = self.running_jobs[job_id]
            os.kill(process.pid, signal.SIGTERM)
            process.join(timeout=5)
            if process.is_alive():
                os.kill(process.pid, signal.SIGKILL)
            del self.running_jobs[job_id]
            return f"Job {job_id} terminated"
        return f"Job {job_id} not found"

    def get_script_part(self, command: str) -> str:
        return command


def get_cluster_manager(dry: bool = False) -> ClusterManager:
    """
    Autodetect the cluster type and return the appropriate ClusterManager.

    Args:
        dry: If True, return LSFManager even if no cluster is detected.

    Returns:
        An instance of the appropriate ClusterManager subclass.
    """
    virtual = os.environ.get("VIRTUAL_CLUSTER", "").lower() in ("true", "1", "yes", "on")
    dry = dry or os.environ.get("DRYRUN_ONLY", "").lower() in ("true", "1", "yes", "on")

    if subprocess.getoutput("command -v bsub"):
        return LSFManager()
    elif subprocess.getoutput("command -v sbatch"):
        return SLURMManager()
    else:
        if dry:
            return LSFManager()
        elif virtual:
            warnings.warn(
                "No cluster management system detected. Using VirtualClusterManager for "
                "local execution. This is not recommended for production use.",
                UserWarning,
                stacklevel=2,
            )
            return VirtualClusterManager()
        else:
            raise RuntimeError("No cluster management system detected.")


class LazyClusterManager:
    """Lazy initialization of the cluster manager."""

    def __init__(self):
        self._instance = None
        # TODO: dry run handling currently not elegant but works for now
        self._dry = False

    def set_dry(self, dry: bool):
        """Set dry run mode."""
        self._dry = dry
        self._instance = None  # Reset instance to force re-initialization

    def __getattr__(self, name):
        if self._instance is None:
            self._instance = get_cluster_manager(dry=self._dry)
        return getattr(self._instance, name)


CLUSTER_MANAGER = LazyClusterManager()


def run_job(command: str, dry: bool) -> str:
    """
    Run a job on the cluster.

    Args:
        command: The command to run.
        dry: If True, perform a dry run without actually submitting the job.

    Returns:
        The job ID as a string, or "dry run" for dry runs.
    """
    # TODO: dry handling currently not elegant but works for now
    env_dry = os.environ.get("DRYRUN_ONLY", "").lower() in ("true", "1", "yes", "on")
    is_dry = dry or env_dry

    if is_dry:
        job_id = "dry run"
        logger.info("Dry run command: %s", command)
        return job_id

    return CLUSTER_MANAGER.run_job(command)


def is_running(job_id: str, dry: bool) -> bool:
    """
    Check if a job is running.

    Args:
        job_id: The ID of the job to check.
        dry: If True, always return False.

    Returns:
        True if the job is running, False otherwise.
    """
    if dry:
        return False
    return CLUSTER_MANAGER.is_running(job_id)


def kill_job(job_id: str, dry: bool) -> str:
    """
    Kill a running job.

    Args:
        job_id: The ID of the job to kill.
        dry: If True, return a message without actually killing the job.

    Returns:
        The output of the kill command or a dry run message.
    """
    if dry:
        return f"Would kill job {job_id}"
    return CLUSTER_MANAGER.kill_job(job_id)


def wait_for_single(job_id: str, dry: bool = False) -> None:
    """
    Wait for a single job to finish on the cluster.

    Args:
        job_id: The ID of the job to wait for.
        job_name: The name of the job.
        dry: If True, skip actual waiting.

    Raises:
        KeyboardInterrupt: If the waiting is interrupted by the user.
    """
    try:
        if not dry:
            sleep(60)
        while is_running(job_id, dry):
            if not dry:
                sleep(60)
    except KeyboardInterrupt as e:
        logger.info("Killing job %s", kill_job(job_id, dry))
        raise KeyboardInterrupt from e


def wait_for_many(job_id_names: Dict[str, str], dry: bool = False) -> None:
    """
    Wait for multiple jobs to finish on the cluster.

    Args:
        job_id_names: A dictionary mapping job IDs to job names.
        dry: If True, skip actual waiting.

    Raises:
        KeyboardInterrupt: If the waiting is interrupted by the user.
    """
    try:
        if not dry:
            print("Jobs launched.. waiting 60s..")
            sleep(60)
        while any(is_running(job_id, dry) for job_id in job_id_names):
            if not dry:
                print("Jobs still running.. waiting 60s..")
                sleep(60)
    except KeyboardInterrupt as e:
        for job_id in job_id_names:
            logger.info("Killing job %s", kill_job(job_id, dry))
        raise KeyboardInterrupt from e


def check_valid_host(blacklist: List[str]) -> None:
    """
    Prevent running on certain blacklisted hosts, e.g., login nodes.

    Args:
        blacklist: A list of blacklisted hostnames or substrings.

    Raises:
        ValueError: If the current host is in the blacklist.
    """
    host = socket.gethostname()
    if any(h in host for h in blacklist):
        raise ValueError(f"This script should not be run from {host}!")


def launch_range(
    start: int,
    end: int,
    ensemble_id: str,
    task_name: str,
    nP: int,
    gpu: str,
    q: str,
    script: str,
    dry: bool,
    kwargs: List[str],
) -> None:
    """
    Launch a range of models.

    Args:
        start: The starting index for the range.
        end: The ending index for the range.
        ensemble_id: The ID of the ensemble.
        task_name: The name of the task.
        nP: The number of processors to use.
        gpu: The GPU configuration.
        q: The queue to submit the job to.
        script: The script to run.
        dry: If True, perform a dry run without actually submitting jobs.
        kwargs: A list of additional keyword arguments for the script.

    Note:
        kwargs is an ordered list of strings, either in the format ["-kw", "val", ...]
        or following hydra syntax, i.e. ["kw=val", ...].
    """
    SCRIPT_PART = "{} {} {}"

    CLUSTER_MANAGER.set_dry(dry)

    job_id_names = {}
    for i in range(start, end):
        kw = kwargs.copy()
        ensemble_and_network_id = f"{ensemble_id:04}/{i:03}"
        assert "_" not in ensemble_and_network_id
        network_dir = results_dir / task_name / ensemble_and_network_id
        if not network_dir.parent.exists():
            network_dir.parent.mkdir(parents=True)
        log_file = (
            network_dir.parent / f"{i:04}_{script.split('/')[-1].split('.')[0]}.log"
        )
        if log_file.exists():
            log_file.unlink()

        kw.extend([f"ensemble_and_network_id={ensemble_and_network_id}"])
        kw.extend([f"task_name={task_name}"])

        LSF_CMD = CLUSTER_MANAGER.get_submit_command(
            f"{task_name}_{ensemble_and_network_id}", nP, log_file, gpu, q
        )
        SCRIPT_CMD = SCRIPT_PART.format(sys.executable, script, " ".join(kw))
        command = LSF_CMD + CLUSTER_MANAGER.get_script_part(SCRIPT_CMD)
        logger.info("Launching command: %s", command)
        job_id = run_job(command, dry)
        job_id_names[job_id] = f"{task_name}_{ensemble_and_network_id}"

    wait_for_many(job_id_names, dry)


def launch_single(
    ensemble_id: str,
    task_name: str,
    nP: int,
    gpu: str,
    q: str,
    script: str,
    dry: bool,
    kwargs: List[str],
) -> None:
    """
    Launch a single job for an ensemble.

    Args:
        ensemble_id: The ID of the ensemble.
        task_name: The name of the task.
        nP: The number of processors to use.
        gpu: The GPU configuration.
        q: The queue to submit the job to.
        script: The script to run.
        dry: If True, perform a dry run without actually submitting the job.
        kwargs: A list of additional keyword arguments for the script.

    Note:
        kwargs is an ordered list of strings, either in the format ["-kw", "val", ...]
        or following hydra syntax, i.e. ["kw=val", ...].
    """
    SCRIPT_PART = "{} {} {}"

    CLUSTER_MANAGER.set_dry(dry)

    job_id_names = {}
    kw = kwargs.copy()
    ensemble_id = f"{ensemble_id:04}"
    assert "_" not in ensemble_id
    ensemble_dir = results_dir / task_name / ensemble_id

    assert ensemble_dir.exists()
    log_file = ensemble_dir / f"{script.split('/')[-1].split('.')[0]}.log"
    if log_file.exists():
        log_file.unlink()

    kw.extend([f"ensemble_id={ensemble_id}"])
    kw.extend([f"task_name={task_name}"])

    LSF_CMD = CLUSTER_MANAGER.get_submit_command(
        f"{task_name}_{ensemble_id}", nP, log_file, gpu, q
    )
    SCRIPT_CMD = SCRIPT_PART.format(sys.executable, script, " ".join(kw))
    command = LSF_CMD + CLUSTER_MANAGER.get_script_part(SCRIPT_CMD)
    logger.info("Launching command: %s", command)
    job_id = run_job(command, dry)
    job_id_names[job_id] = f"{task_name}_{ensemble_id}"

    wait_for_many(job_id_names, dry)



from typing import List, Tuple, Union

import numpy as np
from matplotlib.colors import (
    LinearSegmentedColormap,
    ListedColormap,
    hex2color,
    to_rgba,
)

from flyvis.analysis.visualization.plt_utils import color_labels
from flyvis.utils.groundtruth_utils import polarity

ON = "#af0019"  # red
OFF = "#00b2b2"  # blue

ON_FR = "#c1b933"  # yellow
OFF_FR = "#b140cc"  # violett

# ON_OSI = "#feb24c"  # light orange
# OFF_OSI = "#a1d99b"  # light green
ON_OSI = "#B74F0D"  # reddish like ON
OFF_OSI = "#0089B9"  # blueish like OFF

PD = "#1b9e77"
ND = "#e7298a"

INH = "#0000ff"
EXC = "#FF0000"


def is_hex(color: Union[str, Tuple[float, float, float]]) -> bool:
    """
    Check if the given color is in hexadecimal format.

    Args:
        color: The color to check.

    Returns:
        True if the color is in hexadecimal format, False otherwise.
    """
    return "#" in color


def is_integer_rgb(color: Union[Tuple[float, float, float], List[float]]) -> bool:
    """
    Check if the given color is in integer RGB format (0-255).

    Args:
        color: The color to check.

    Returns:
        True if the color is in integer RGB format, False otherwise.
    """
    try:
        return any([c > 1 for c in color])
    except TypeError:
        return False


def single_color_cmap(color: Union[str, Tuple[float, float, float]]) -> ListedColormap:
    """
    Create a single color colormap.

    Args:
        color: The color to use for the colormap.

    Returns:
        A ListedColormap object with the specified color.
    """
    if is_hex(color):
        color = to_rgba(color)
    elif is_integer_rgb(color):
        color = np.array(color) / 255
    return ListedColormap(color)


def color_to_cmap(
    end_color: str,
    start_color: str = "#FFFFFF",
    name: str = "custom_cmap",
    N: int = 256,
) -> LinearSegmentedColormap:
    """
    Create a colormap from start and end colors.

    Args:
        end_color: The end color of the colormap.
        start_color: The start color of the colormap.
        name: The name of the colormap.
        N: The number of color segments.

    Returns:
        A LinearSegmentedColormap object.
    """
    return LinearSegmentedColormap.from_list(
        name,
        [hex2color(start_color), hex2color(end_color)],
        N=N,
    )


def get_alpha_colormap(
    saturated_color: Union[str, Tuple[float, float, float]], number_of_shades: int
) -> ListedColormap:
    """
    Create a colormap from a color and a number of shades.

    Args:
        saturated_color: The base color for the colormap.
        number_of_shades: The number of shades to create.

    Returns:
        A ListedColormap object with varying alpha values.
    """
    if is_hex(saturated_color):
        rgba = [*hex2color(saturated_color)[:3], 0]
    elif is_integer_rgb(saturated_color):
        rgba = [*list(np.array(saturated_color) / 255.0), 0]

    colors = []
    alphas = np.linspace(1 / number_of_shades, 1, number_of_shades)[::-1]
    for alpha in alphas:
        rgba[-1] = alpha
        colors.append(rgba.copy())

    return ListedColormap(colors)


def adapt_color_alpha(
    color: Union[str, Tuple[float, float, float], Tuple[float, float, float, float]],
    alpha: float,
) -> Tuple[float, float, float, float]:
    """
    Transform a color specification to RGBA and adapt the alpha value.

    Args:
        color: Color specification in various formats: hex string, RGB tuple, or
            RGBA tuple.
        alpha: New alpha value to be applied.

    Returns:
        The adapted color in RGBA format.
    """
    color_rgb = to_rgba(color)
    r, g, b, _ = color_rgb
    return r, g, b, alpha


def flash_response_color_labels(ax):
    """
    Apply color labels for ON and OFF flash responses.

    Args:
        ax: The matplotlib axis to apply the labels to.

    Returns:
        The modified matplotlib axis.
    """
    on = [key for key, value in polarity.items() if value == 1]
    off = [key for key, value in polarity.items() if value == -1]
    color_labels(on, ON_FR, ax)
    color_labels(off, OFF_FR, ax)
    return ax


single_blue_cmap = single_color_cmap("#2c7fb8")
single_orange_cmap = single_color_cmap("#d95f0e")

cell_type_colors = {
    "R1": "#000000",
    "R2": "#000000",
    "R3": "#000000",
    "R4": "#000000",
    "R5": "#000000",
    "R6": "#000000",
    "R7": "#000000",
    "R8": "#000000",
    "L1": "#df6e26",
    "L2": "#df6e26",
    "L3": "#df6e26",
    "L4": "#df6e26",
    "L5": "#df6e26",
    "Lawf1": "#df6e26",
    "Lawf2": "#df6e26",
    "Am": "#000000",
    "C2": "#000000",
    "C3": "#000000",
    "CT1(Lo1)": "#000000",
    "CT1(M10)": "#000000",
    "Mi1": "#5db5ac",
    "Mi2": "#5db5ac",
    "Mi3": "#5db5ac",
    "Mi4": "#5db5ac",
    "Mi9": "#5db5ac",
    "Mi10": "#5db5ac",
    "Mi11": "#5db5ac",
    "Mi12": "#5db5ac",
    "Mi13": "#5db5ac",
    "Mi14": "#5db5ac",
    "Mi15": "#5db5ac",
    "T4a": "#ed3533",
    "T4b": "#ed3533",
    "T4c": "#ed3533",
    "T4d": "#ed3533",
    "T1": "#ed3533",
    "T2": "#ed3533",
    "T2a": "#ed3533",
    "T3": "#ed3533",
    "T5a": "#ed3533",
    "T5b": "#ed3533",
    "T5c": "#ed3533",
    "T5d": "#ed3533",
    "Tm1": "#a454a0",
    "Tm2": "#a454a0",
    "Tm3": "#a454a0",
    "Tm4": "#a454a0",
    "Tm5Y": "#a454a0",
    "Tm5a": "#a454a0",
    "Tm5b": "#a454a0",
    "Tm5c": "#a454a0",
    "Tm9": "#a454a0",
    "Tm16": "#a454a0",
    "Tm20": "#a454a0",
    "Tm28": "#a454a0",
    "Tm30": "#a454a0",
    "TmY3": "#a454a0",
    "TmY4": "#a454a0",
    "TmY5a": "#a454a0",
    "TmY9": "#a454a0",
    "TmY10": "#a454a0",
    "TmY13": "#a454a0",
    "TmY14": "#a454a0",
    "TmY15": "#a454a0",
    "TmY18": "#a454a0",
}


def truncate_colormap(
    cmap: Union[LinearSegmentedColormap, ListedColormap],
    minval: float = 0.0,
    maxval: float = 1.0,
    n: int = 100,
) -> LinearSegmentedColormap:
    """
    Truncate a colormap to a specific range.

    Args:
        cmap: The colormap to truncate.
        minval: The minimum value of the new range.
        maxval: The maximum value of the new range.
        n: The number of color segments in the new colormap.

    Returns:
        A new LinearSegmentedColormap with the truncated range.
    """
    new_cmap = LinearSegmentedColormap.from_list(
        "trunc({n},{a:.2f},{b:.2f})".format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, max(n, 2))),
    )
    return new_cmap.resampled(max(n, 2))


class cmap_iter:
    """
    An iterator for colormap colors.

    Attributes:
        i: The current index.
        cmap: The colormap to iterate over.
        stop: The number of colors in the colormap.
    """

    def __init__(self, cmap: Union[LinearSegmentedColormap, ListedColormap]):
        """
        Initialize the cmap_iter.

        Args:
            cmap: The colormap to iterate over.
        """
        self.i: int = 0
        self.cmap = cmap
        self.stop: int = cmap.N

    def __next__(self) -> Tuple[float, float, float, float]:
        """
        Get the next color from the colormap.

        Returns:
            The next color as an RGBA tuple.
        """
        if self.i < self.stop:
            self.i += 1
            return self.cmap(self.i - 1)

    def _repr_html_(self) -> str:
        """
        Return the HTML representation of the colormap.

        Returns:
            The HTML representation of the colormap.
        """
        return self.cmap._repr_html_()



from functools import wraps
from typing import Any, Callable, TypeVar

T = TypeVar("T")


def context_aware_cache(
    func: Callable[..., T] = None, context: Callable[[Any], Any] = lambda self: None
) -> Callable[..., T]:
    """
    Decorator to cache the result of a method based on its arguments and context.

    Args:
        func: The function to be decorated.
        context: A function that returns the context for caching.

    Returns:
        A wrapped function that implements caching based on arguments and context.

    Example:
        ```python
        class MyClass:
            def __init__(self):
                self.cache = {}

            @context_aware_cache(context=lambda self: self.some_attribute)
            def my_method(self, arg1, arg2):
                # Method implementation
                pass
        ```
    """
    if func is None:

        def decorator(f: Callable[..., T]) -> Callable[..., T]:
            @wraps(f)
            def wrapper(self: Any, *args: Any, **kwargs: Any) -> T:
                context_key = make_hashable(context(self))
                cache_key = hash(make_hashable((f.__name__, args, kwargs, context_key)))
                if cache_key in self.cache:
                    return self.cache[cache_key]
                result = f(self, *args, **kwargs)
                self.cache[cache_key] = result
                return result

            return wrapper

        return decorator
    else:

        @wraps(func)
        def wrapper(self: Any, *args: Any, **kwargs: Any) -> T:
            context_key = make_hashable(context(self))
            cache_key = hash(make_hashable((func.__name__, args, kwargs, context_key)))
            if cache_key in self.cache:
                return self.cache[cache_key]
            result = func(self, *args, **kwargs)
            self.cache[cache_key] = result
            return result

        return wrapper


def make_hashable(obj: Any) -> Any:
    """Recursively converts an object into a hashable type."""
    if isinstance(obj, (int, float, str, bool, type(None))):
        return obj
    elif isinstance(obj, (list, set)):
        try:
            # Try direct sorting first
            return tuple(make_hashable(e) for e in sorted(obj))
        except TypeError:
            # Fall back to sorting by hash
            return tuple(
                make_hashable(e)
                for e in sorted(obj, key=lambda x: hash(make_hashable(x)))
            )
    elif isinstance(obj, dict):
        try:
            # Try direct sorting of keys first
            return tuple(sorted((k, make_hashable(v)) for k, v in obj.items()))
        except TypeError:
            # Fall back to sorting by hash of keys
            return tuple(
                sorted(
                    ((k, make_hashable(v)) for k, v in obj.items()),
                    key=lambda x: hash(make_hashable(x[0])),
                )
            )
    elif isinstance(obj, (tuple, frozenset)):
        return tuple(make_hashable(e) for e in obj)
    elif isinstance(obj, slice):
        return (obj.start, obj.stop, obj.step)
    else:
        # For other types, try to get a consistent string representation
        return f"{obj.__class__.__module__}.{obj.__class__.__name__}:{str(obj)}"
